@@ -1,6 +1,6 @@
 /*
-    roxterm - GTK+ 2.0 terminal emulator with tabs
-    Copyright (C) 2004 Tony Houghton <h@realh.co.uk>
+    roxterm - VTE/GTK terminal emulator with tabs
+    Copyright (C) 2004-2011 Tony Houghton <h@realh.co.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -87,6 +87,7 @@ static gboolean global_options_show_usage(const gchar *option_name,
       "    [--directory=DIRECTORY|-d DIRECTORY] [--disable-sm]\n"
       "    [--show-menubar] [--hide-menubar]\n"
       "    [--role=ROLE]\n"
+      "    [--no-geometry]\n"
       "    [-e|--execute COMMAND]\n");
     exit(0);
     return TRUE;
@@ -257,6 +258,12 @@ static GOptionEntry global_g_options[] = {
         G_OPTION_ARG_STRING, &global_options_restart_session_id, NULL, NULL },
     { "clone-session-id", 0, G_OPTION_FLAG_HIDDEN,
         G_OPTION_ARG_STRING, &global_options_clone_session_id, NULL, NULL },
+#if GTK_CHECK_VERSION(3, 0, 0)
+    { "no-geometry", 0, G_OPTION_FLAG_IN_MAIN | G_OPTION_FLAG_NO_ARG,
+        G_OPTION_ARG_CALLBACK, global_options_set_bool,
+        N_("Don't set window geometry hints. This is a workaround for "
+        "<https://bugzilla.gnome.org/show_bug.cgi?id=649680>"), NULL },
+#endif
     { "execute", 'e', G_OPTION_FLAG_IN_MAIN | G_OPTION_FLAG_NO_ARG,
         G_OPTION_ARG_CALLBACK, global_options_swallow_execute,
         N_("Execute remainder of command line inside the\n"
