@@ -1,6 +1,6 @@
 /*
-    roxterm - GTK+ 2.0 terminal emulator with tabs
-    Copyright (C) 2004 Tony Houghton <h@realh.co.uk>
+    roxterm - VTE/GTK terminal emulator with tabs
+    Copyright (C) 2004-2011 Tony Houghton <h@realh.co.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,14 +29,20 @@
 static char const *menutree_labels[MENUTREE_NUM_IDS];
 static gboolean filled_labels = FALSE;
 
-inline static GtkWidget *menutree_append_new_item_with_mnemonic(GtkMenuShell *
-    shell, const char *label, const char *stock)
+static GtkWidget *menutree_append_new_item_with_mnemonic(
+        GtkMenuShell *shell,
+        const char *label, const char *stock, gboolean separate_label)
 {
     GtkWidget *item;
         
     if (stock)
     {
         item = gtk_image_menu_item_new_from_stock(stock, NULL);
+        if (separate_label)
+        {
+            gtk_menu_item_set_label(GTK_MENU_ITEM(item), label);
+            gtk_menu_item_set_use_underline(GTK_MENU_ITEM(item), TRUE);
+        }
     }
     else
     {
@@ -82,6 +88,7 @@ static void menutree_build_shell_ap(MenuTree *menu_tree, GtkMenuShell *shell,
         }
         else
         {
+            gboolean separate_label = FALSE;
             const char *stock = NULL;
 
             if (!filled_labels &&
@@ -93,12 +100,37 @@ static void menutree_build_shell_ap(MenuTree *menu_tree, GtkMenuShell *shell,
                 label = strip_underscore(label);
             switch (id)
             {
+                case MENUTREE_FILE_NEW_WINDOW:
+                case MENUTREE_FILE_NEW_TAB:
+                    stock = GTK_STOCK_NEW;
+                    separate_label = TRUE;
+                    break;
+                case MENUTREE_FILE_CLOSE_WINDOW:
+                case MENUTREE_FILE_CLOSE_TAB:
+                case MENUTREE_TABS_CLOSE_TAB:
+                    stock = GTK_STOCK_CLOSE;
+                    separate_label = TRUE;
+                    break;
                 case MENUTREE_EDIT_COPY:
                     stock = GTK_STOCK_COPY;
                     break;
                 case MENUTREE_EDIT_PASTE:
                     stock = GTK_STOCK_PASTE;
                     break;
+                case MENUTREE_EDIT_SET_WINDOW_TITLE:
+                case MENUTREE_TABS_NAME_TAB:
+                    stock = GTK_STOCK_EDIT;
+                    separate_label = TRUE;
+                    break;
+                case MENUTREE_EDIT_RESET:
+                case MENUTREE_EDIT_RESET_AND_CLEAR:
+                    stock = GTK_STOCK_CLEAR;
+                    separate_label = TRUE;
+                    break;
+                case MENUTREE_EDIT_RESPAWN:
+                    stock = GTK_STOCK_REFRESH;
+                    separate_label = TRUE;
+                    break;
                 case MENUTREE_VIEW_ZOOM_IN:
                     stock = GTK_STOCK_ZOOM_IN;
                     break;
@@ -108,6 +140,22 @@ static void menutree_build_shell_ap(MenuTree *menu_tree, GtkMenuShell *shell,
                 case MENUTREE_VIEW_ZOOM_NORM:
                     stock = GTK_STOCK_ZOOM_100;
                     break;
+#ifdef HAVE_VTE_TERMINAL_SEARCH_SET_GREGEX
+                case MENUTREE_SEARCH_FIND:
+                case MENUTREE_SEARCH_FIND_NEXT:
+                case MENUTREE_SEARCH_FIND_PREVIOUS:
+                    stock = GTK_STOCK_FIND;
+                    separate_label = TRUE;
+                    break;
+#endif
+                case MENUTREE_PREFERENCES_EDIT_CURRENT_PROFILE:
+                    stock = GTK_STOCK_PREFERENCES;
+                    separate_label = TRUE;
+                    break;
+                case MENUTREE_PREFERENCES_EDIT_CURRENT_COLOUR_SCHEME:
+                    stock = GTK_STOCK_SELECT_COLOR;
+                    separate_label = TRUE;
+                    break;
                 case MENUTREE_HELP_SHOW_MANUAL:
                     stock = GTK_STOCK_HELP;
                     break;
@@ -128,7 +176,7 @@ static void menutree_build_shell_ap(MenuTree *menu_tree, GtkMenuShell *shell,
                     break;
                 default:
                     item = menutree_append_new_item_with_mnemonic(shell,
-                            label, stock);
+                            label, stock, separate_label);
                     break;
             }
             if (menu_tree->disable_shortcuts)
@@ -152,13 +200,20 @@ static void menutree_build_shell(MenuTree *menu_tree, GtkMenuShell * shell, ...)
 #define URI_MENU_ITEMS \
         _("_Open URL in browser"), MENUTREE_OPEN_IN_BROWSER, \
         _("_Send email"), MENUTREE_OPEN_IN_MAILER, \
+        _("_Open file/directory in filer"), MENUTREE_OPEN_IN_FILER, \
         _("_Copy address to clipboard"), MENUTREE_COPY_URI, \
         _("_"), MENUTREE_URI_SEPARATOR
         
+#ifdef HAVE_VTE_TERMINAL_SEARCH_SET_GREGEX
+#define MENUTREE_SEARCH_ITEM _("_Search"), MENUTREE_SEARCH,
+#else
+#define MENUTREE_SEARCH_ITEM
+#endif
 #define TOP_LEVEL_MENU_ITEMS \
         _("_File"), MENUTREE_FILE, \
         _("_Edit"), MENUTREE_EDIT, \
         _("_View"), MENUTREE_VIEW, \
+        MENUTREE_SEARCH_ITEM \
         _("_Preferences"), MENUTREE_PREFERENCES, \
         _("Ta_bs"), MENUTREE_TABS, \
         _("_Help"), MENUTREE_HELP
@@ -191,7 +246,7 @@ GtkMenu *menutree_submenu_from_id(MenuTree *mtree, MenuTreeID id)
 static char *get_accel_path(Options *shortcuts, const char *branch_name)
 {
     return g_strjoin("/", ACCEL_PATH,
-            (const char *) options_get_data(shortcuts), branch_name, NULL);
+            shortcuts_get_index_str(shortcuts), branch_name, NULL);
 }
 
 static void
@@ -263,8 +318,12 @@ void menutree_apply_shortcuts(MenuTree *tree, Options *shortcuts)
     menutree_set_accel_path_for_submenu(tree, MENUTREE_FILE, "File");
     menutree_set_accel_path_for_submenu(tree, MENUTREE_EDIT, "Edit");
     menutree_set_accel_path_for_submenu(tree, MENUTREE_VIEW, "View");
+#ifdef HAVE_VTE_TERMINAL_SEARCH_SET_GREGEX
+    menutree_set_accel_path_for_submenu(tree, MENUTREE_SEARCH, "Search");
+#endif
     menutree_set_accel_path_for_submenu(tree, MENUTREE_PREFERENCES,
             "Preferences");
+    menutree_set_accel_path_for_submenu(tree, MENUTREE_HELP, "Help");
     
     submenu = GTK_MENU(tree->new_win_profiles_menu);
     if (submenu)
@@ -296,6 +355,12 @@ void menutree_apply_shortcuts(MenuTree *tree, Options *shortcuts)
     menutree_set_accel_path_for_item(tree,
             MENUTREE_FILE_NEW_TAB_WITH_PROFILE_HEADER,
             "File/New Tab With Profile/Profiles");
+    menutree_set_accel_path_for_item(tree, MENUTREE_TABS_DETACH_TAB,
+            "Tabs/Detach Tab");
+    menutree_set_accel_path_for_item(tree, MENUTREE_TABS_CLOSE_TAB,
+            "Tabs/Close Tab");
+    menutree_set_accel_path_for_item(tree, MENUTREE_TABS_CLOSE_OTHER_TABS,
+            "Tabs/Close Other Tabs");
     menutree_set_accel_path_for_item(tree, MENUTREE_TABS_NAME_TAB,
             "Tabs/Name Tab...");
     menutree_set_accel_path_for_item(tree, MENUTREE_TABS_NEXT_TAB,
@@ -549,7 +614,7 @@ static void menutree_build(MenuTree *menu_tree, Options *shortcuts,
     submenu = gtk_menu_new();
     menutree_build_shell(menu_tree, GTK_MENU_SHELL(submenu),
         SHOW_MENU_BAR_ITEM,
-        _("Always Show _Tab Bar"), MENUTREE_VIEW_SHOW_TAB_BAR,
+        _("_Always Show Tab Bar"), MENUTREE_VIEW_SHOW_TAB_BAR,
         _("_Full Screen"), MENUTREE_VIEW_FULLSCREEN,
         "_", MENUTREE_NULL_ID,
         _("Zoom _In"), MENUTREE_VIEW_ZOOM_IN,
@@ -558,10 +623,25 @@ static void menutree_build(MenuTree *menu_tree, Options *shortcuts,
         "_", MENUTREE_NULL_ID,
         _("Scroll _Up One Line"), MENUTREE_VIEW_SCROLL_UP,
         _("Scroll _Down One Line"), MENUTREE_VIEW_SCROLL_DOWN,
+        _("Scroll Up One _Page"), MENUTREE_VIEW_SCROLL_PAGE_UP,
+        _("Scroll Do_wn One Page"), MENUTREE_VIEW_SCROLL_PAGE_DOWN,
+        _("Scroll To _Top"), MENUTREE_VIEW_SCROLL_TO_TOP,
+        _("Scroll To _Bottom"), MENUTREE_VIEW_SCROLL_TO_BOTTOM,
         NULL);
     gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_tree->item_widgets
             [MENUTREE_VIEW]), submenu);
 
+#ifdef HAVE_VTE_TERMINAL_SEARCH_SET_GREGEX
+    submenu = gtk_menu_new();
+    menutree_build_shell(menu_tree, GTK_MENU_SHELL(submenu),
+        _("_Find..."), MENUTREE_SEARCH_FIND,
+        _("Find _Next"), MENUTREE_SEARCH_FIND_NEXT,
+        _("Find _Previous"), MENUTREE_SEARCH_FIND_PREVIOUS,
+        NULL);
+    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_tree->item_widgets
+            [MENUTREE_SEARCH]), submenu);
+#endif
+
     submenu = gtk_menu_new();
     menutree_build_shell(menu_tree, GTK_MENU_SHELL(submenu),
         _("Select _Profile"), MENUTREE_PREFERENCES_SELECT_PROFILE,
@@ -581,6 +661,9 @@ static void menutree_build(MenuTree *menu_tree, Options *shortcuts,
 
     submenu = gtk_menu_new();
     menutree_build_shell(menu_tree, GTK_MENU_SHELL(submenu),
+        _("_Detach Tab"), MENUTREE_TABS_DETACH_TAB,
+        _("_Close Tab"), MENUTREE_TABS_CLOSE_TAB,
+        _("Close _Other Tabs"), MENUTREE_TABS_CLOSE_OTHER_TABS,
         _("Na_me Tab..."), MENUTREE_TABS_NAME_TAB,
         "_", MENUTREE_NULL_ID,
         _("_Previous Tab"), MENUTREE_TABS_PREVIOUS_TAB,
@@ -622,7 +705,7 @@ static void menutree_destroy(MenuTree * tree)
     g_free(tree);
 }
 
-static void menutree_destroy_widget_handler(GtkObject * obj, MenuTree * tree)
+static void menutree_destroy_widget_handler(GObject * obj, MenuTree * tree)
 {
     (void) obj;
 
@@ -648,8 +731,8 @@ static MenuTree *menutree_new_common(Options *shortcuts,
 
     builder(tree, shortcuts, menu_type);
 
-    g_signal_connect(GTK_OBJECT(tree->top_level),
-        "destroy", G_CALLBACK(menutree_destroy_widget_handler), tree);
+    g_signal_connect(tree->top_level,
+            "destroy", G_CALLBACK(menutree_destroy_widget_handler), tree);
 
     gtk_widget_show_all(tree->top_level);
     return tree;
