@@ -1,6 +1,6 @@
 /*
-    roxterm - GTK+ 2.0 terminal emulator with tabs
-    Copyright (C) 2004 Tony Houghton <h@realh.co.uk>
+    roxterm - VTE/GTK terminal emulator with tabs
+    Copyright (C) 2004-2011 Tony Houghton <h@realh.co.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,13 +22,21 @@
 #include "globalopts.h"
 #include "menutree.h"
 #include "multitab.h"
+#include "multitab-close-button.h"
+#include "multitab-label.h"
 #include "shortcuts.h"
 #if HAVE_COMPOSITE
 #include <gdk/gdkx.h>
 #include "x11support.h"
 #endif
 
-#define HORIZ_TAB_WIDTH 120
+#define HORIZ_TAB_WIDTH_CHARS 16
+
+#if GTK_CHECK_VERSION(3,0,0)
+#define HAVE_COLORMAP 0
+#else
+#define HAVE_COLORMAP 1
+#endif
 
 struct MultiTab {
     MultiWin *parent;
@@ -40,17 +48,11 @@ struct MultiTab {
     GtkWidget *active_widget;
     gpointer user_data;
     GtkWidget *label;
-    GtkWidget *label_bg;
-    GtkWidget *menu_label;
     GtkAdjustment *adjustment;
     double scroll_step;
-    gboolean attention;
-    guint attention_tag;
     GtkWidget *close_button;
     GtkWidget *label_box;
-    void (*label_packer)(GtkBox *, GtkWidget *, gboolean, gboolean, guint);
     const char *status_stock;
-    GtkWidget *image;
     GtkWidget *rename_dialog;
     gboolean postponed_free;
     gboolean old_win_destroyed;
@@ -97,6 +99,9 @@ struct MultiWin {
     char *display_name;
     gboolean title_template_locked;
     gboolean clear_demands_attention;
+#if MULTITAB_LABEL_GTK3_SIZE_KLUDGE
+    int best_tab_width;
+#endif
 };
 
 static double multi_win_zoom_factors[] = {
@@ -134,6 +139,7 @@ MultiWinGetDisableMenuShortcuts multi_win_get_disable_menu_shortcuts;
 static MultiWinInitialTabs multi_win_initial_tabs;
 static MultiWinDeleteHandler multi_win_delete_handler;
 static MultiTabGetShowCloseButton multi_tab_get_show_close_button;
+static MultiTabGetNewTabAdjacent multi_tab_get_new_tab_adjacent;
 
 static gboolean multi_win_notify_tab_removed(MultiWin *, MultiTab *);
 
@@ -155,12 +161,21 @@ static void multi_win_restore_size(MultiWin *win)
     MultiTab *tab = win->current_tab;
     int w = -1;
     int h = -1;
+#if !GTK_CHECK_VERSION(3, 0, 0)
     GtkRequisition win_rq, term_rq;
+#endif
     
-    if (!tab || !tab->active_widget || !GTK_WIDGET_REALIZED(tab->active_widget))
+    if (!tab || !tab->active_widget ||
+            !gtk_widget_get_realized(tab->active_widget))
+    {
         return;
+    }
     if (multi_win_is_maximised(win) || multi_win_is_fullscreen(win))
         return;
+#if GTK_CHECK_VERSION(3, 0, 0)
+    multi_win_size_func(tab->user_data, FALSE, &w, &h);
+    gtk_window_resize_to_geometry(GTK_WINDOW(win->gtkwin), w, h);
+#else
     /* Find new difference between window and term widget (padding) */
     gtk_widget_size_request(win->gtkwin, &win_rq);
     gtk_widget_size_request(tab->active_widget, &term_rq);
@@ -174,6 +189,7 @@ static void multi_win_restore_size(MultiWin *win)
     gtk_window_resize(GTK_WINDOW(win->gtkwin),
             w + win_rq.width - term_rq.width,
             h + win_rq.height - term_rq.height);
+#endif
 }
 
 MultiTab *multi_tab_get_from_widget(GtkWidget *widget)
@@ -202,7 +218,8 @@ void multi_tab_init(MultiTabFiller filler, MultiTabDestructor destructor,
     MultiWinGetDisableMenuShortcuts get_disable_menu_shortcuts,
     MultiWinInitialTabs initial_tabs,
     MultiWinDeleteHandler delete_handler,
-    MultiTabGetShowCloseButton get_show_close_button
+    MultiTabGetShowCloseButton get_show_close_button,
+    MultiTabGetNewTabAdjacent get_new_tab_adjacent
     )
 {
     multi_tab_filler = filler;
@@ -217,6 +234,7 @@ void multi_tab_init(MultiTabFiller filler, MultiTabDestructor destructor,
     multi_win_initial_tabs = initial_tabs;
     multi_win_delete_handler = delete_handler;
     multi_tab_get_show_close_button = get_show_close_button;
+    multi_tab_get_new_tab_adjacent = get_new_tab_adjacent;
 }
 
 MultiTab *multi_tab_new(MultiWin * parent, gpointer user_data_template)
@@ -229,7 +247,8 @@ MultiTab *multi_tab_new(MultiWin * parent, gpointer user_data_template)
         &tab->user_data, &tab->active_widget, &tab->adjustment);
     g_object_set_data(G_OBJECT(tab->widget), "roxterm_tab", tab);
 
-    if (parent->current_tab)
+    if (parent->current_tab &&
+            multi_tab_get_new_tab_adjacent(user_data_template))
     {
         pos = multi_tab_get_page_num(parent->current_tab) + 1;
         if (pos == parent->ntabs)
@@ -309,11 +328,23 @@ const char *multi_tab_get_display_name(MultiTab *tab)
     return tab->parent ? tab->parent->display_name : NULL;
 }
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+static gboolean multi_win_clear_geometry_hints(GtkWindow *w)
+{
+    gtk_window_set_geometry_hints(w, NULL, NULL, 0);
+    return FALSE;
+}
+#endif
+
 void multi_win_set_geometry_hints(MultiWin *win, GtkWidget *child,
     GdkGeometry *geometry, GdkWindowHints geom_mask)
 {
     gtk_window_set_geometry_hints(GTK_WINDOW(win->gtkwin), child,
         geometry, geom_mask);
+#if GTK_CHECK_VERSION(3, 0, 0)
+    if (global_options_lookup_int_with_default("no-geometry", FALSE))
+        g_idle_add((GSourceFunc) multi_win_clear_geometry_hints, win->gtkwin);
+#endif
 }
 
 static void multi_win_set_geometry_hints_for_tab(MultiWin * win, MultiTab * tab)
@@ -425,14 +456,7 @@ static void multi_tab_set_full_window_title(MultiTab * tab,
 
     if (tab->label)
     {
-        char *menu_text = g_strdup_printf("%d. %s",
-                gtk_notebook_page_num(GTK_NOTEBOOK(win->notebook),
-                        tab->widget) + 1,
-                actual_title);
-        
-        gtk_label_set_text(GTK_LABEL(tab->label), actual_title);
-        gtk_label_set_text(GTK_LABEL(tab->menu_label), menu_text);
-        g_free(menu_text);
+        multitab_label_set_text(MULTITAB_LABEL(tab->label), actual_title);
     }
     if (win)
     {
@@ -534,6 +558,7 @@ MultiTab *multi_tab_get_tab_under_pointer(int x, int y)
     GList *tab_node;
     GtkPositionType tab_pos;
     gboolean vert_tabs;
+    GtkAllocation allocation;
 
     if (!win)
         return NULL;
@@ -548,13 +573,14 @@ MultiTab *multi_tab_get_tab_under_pointer(int x, int y)
         GtkWidget *label = gtk_notebook_get_tab_label(
                 GTK_NOTEBOOK(win->notebook), tab->widget);
 
-        if (!GTK_WIDGET_MAPPED(label))
+        if (!gtk_widget_get_mapped(label))
             continue;
-        gdk_window_get_origin(label->window, &x0, &y0);
-        x0 += label->allocation.x;
-        y0 += label->allocation.y;
-        x1 = x0 + label->allocation.width;
-        y1 = y0 + label->allocation.height;
+        gdk_window_get_origin(gtk_widget_get_window(label), &x0, &y0);
+        gtk_widget_get_allocation(label, &allocation);
+        x0 += allocation.x;
+        y0 += allocation.y;
+        x1 = x0 + allocation.width;
+        y1 = y0 + allocation.height;
         if (vert_tabs && y >= y0 && y < y1)
             return tab;
         else if (!vert_tabs && x >= x0 && x < x1)
@@ -623,9 +649,18 @@ inline static void multi_win_shade_for_next_and_previous_tab(MultiWin * win)
     multi_win_shade_for_next_tab(win);
 }
 
-inline static void multi_win_shade_menus_for_tabs(MultiWin * win)
+static void multi_win_shade_for_single_tab(MultiWin * win)
+{
+    multi_win_shade_item_in_both_menus(win, MENUTREE_TABS_CLOSE_OTHER_TABS,
+            win->ntabs <= 1);
+    multi_win_shade_item_in_both_menus(win, MENUTREE_TABS_DETACH_TAB,
+            win->ntabs <= 1);
+}
+
+static void multi_win_shade_menus_for_tabs(MultiWin * win)
 {
     multi_win_shade_for_next_and_previous_tab(win);
+    multi_win_shade_for_single_tab(win);
 }
 
 static void multi_tab_remove_menutree_items(MultiWin * win, MultiTab * tab)
@@ -710,51 +745,58 @@ void multi_tab_move_to_new_window(MultiWin *win, MultiTab *tab, int position)
     g_object_unref(tab->widget);
 }
 
-static gboolean multi_tab_toggle_attention(gpointer data)
-{
-    MultiTab *tab = data;
-    
-    gtk_event_box_set_visible_window(GTK_EVENT_BOX(tab->label_bg),
-            tab->attention = !tab->attention);
-    return TRUE;
-}
-
 void multi_tab_draw_attention(MultiTab *tab)
 {
-    multi_tab_cancel_attention(tab);
-    multi_tab_toggle_attention(tab);
-    tab->attention_tag = g_timeout_add(500, multi_tab_toggle_attention, tab);
+    multitab_label_draw_attention(MULTITAB_LABEL(tab->label));
 }
 
 void multi_tab_cancel_attention(MultiTab *tab)
 {
-    if (tab->attention_tag)
-    {
-        g_source_remove(tab->attention_tag);
-        tab->attention_tag = 0;
-    }
-    else
-    {
-        return;
-    }
-    if (tab->attention)
-    {
-        multi_tab_toggle_attention(tab);
-        tab->attention = FALSE;
-    }
+    multitab_label_cancel_attention(MULTITAB_LABEL(tab->label));
 }
 
-#if 0
-/* Force tab width to something "nice". Unfortunately we can't
- * easily find the window's size when we're adding the first tab. */
-static void multi_tab_set_single_size(MultiTab *tab)
+static void multi_tab_pack_for_horizontal(MultiTab *tab, GtkContainer *nb)
 {
-    g_object_set(G_OBJECT(tab->parent->notebook), "homogeneous", FALSE, NULL);
-    gtk_container_child_set(GTK_CONTAINER(tab->parent->notebook), tab->widget,
-            "tab-expand", FALSE, "tab-fill", FALSE, NULL);
-    gtk_widget_set_size_request(tab->label, 240, -1);
+    gtk_container_child_set(nb, tab->widget,
+            "tab-expand", FALSE, "tab-fill", TRUE, NULL);
+    multitab_label_set_fixed_width(MULTITAB_LABEL(tab->label),
+            HORIZ_TAB_WIDTH_CHARS);
 }
+
+static void multi_win_pack_for_single_tab(MultiWin *win)
+{
+    MultiTab *tab = win->tabs->data;
+    
+#if !GTK_CHECK_VERSION(3, 0, 0)
+    g_object_set(G_OBJECT(win->notebook), "homogeneous", FALSE, NULL);
 #endif
+    gtk_container_child_set(GTK_CONTAINER(win->notebook), tab->widget,
+            "tab-expand", FALSE, "tab-fill", TRUE, NULL);
+    multitab_label_set_single(MULTITAB_LABEL(tab->label), TRUE);
+    multitab_label_set_fixed_width(MULTITAB_LABEL(tab->label), -1);
+}
+
+static void multi_tab_pack_for_multiple(MultiTab *tab, GtkContainer *nb)
+{
+    gtk_container_child_set(nb, tab->widget,
+            "tab-expand", TRUE, "tab-fill", TRUE, NULL);
+    multitab_label_set_single(MULTITAB_LABEL(tab->label), FALSE);
+    multitab_label_set_fixed_width(MULTITAB_LABEL(tab->label), -1);
+}
+
+static void multi_win_pack_for_multiple_tabs(MultiWin *win)
+{
+    GList *link;
+    GtkContainer *nb = GTK_CONTAINER(win->notebook);
+    
+#if !GTK_CHECK_VERSION(3, 0, 0)
+    g_object_set(G_OBJECT(win->notebook), "homogeneous", TRUE, NULL);
+#endif
+    for (link = win->tabs; link; link = g_list_next(link))
+    {
+        multi_tab_pack_for_multiple(link->data, nb);
+    }
+}
 
 static void multi_win_set_full_title(MultiWin *win,
         const char *template, const char *title)
@@ -780,8 +822,9 @@ void multi_win_set_title(MultiWin *win, const char *title)
 
 static void multi_win_set_icon_title(MultiWin *win, const char *title)
 {
-    if (win->gtkwin->window)
-        gdk_window_set_icon_name(win->gtkwin->window, title);
+    GdkWindow *w = gtk_widget_get_window(win->gtkwin);
+    if (w)
+        gdk_window_set_icon_name(w, title);
 }
 
 void multi_win_select_tab(MultiWin * win, MultiTab * tab)
@@ -800,7 +843,8 @@ void multi_win_select_tab(MultiWin * win, MultiTab * tab)
     {
         char *title = multi_tab_get_full_window_title(tab);
         
-        multi_tab_cancel_attention(tab);
+        if (tab->label)
+            multitab_label_cancel_attention(MULTITAB_LABEL(tab->label));
         multi_tab_set_status_stock(tab, NULL);
         win->user_data_template = tab->user_data;
         gtk_notebook_set_current_page(GTK_NOTEBOOK(win->notebook),
@@ -811,7 +855,7 @@ void multi_win_select_tab(MultiWin * win, MultiTab * tab)
         multi_win_set_icon_title(win, tab->icon_title);
         menutree_select_tab(win->popup_menu, tab->popup_menu_item);
         menutree_select_tab(win->menu_bar, tab->menu_bar_item);
-        if (GTK_WIDGET_REALIZED(tab->active_widget))
+        if (gtk_widget_get_realized(tab->active_widget))
         {
             if (win->tab_selection_handler)
                 win->tab_selection_handler(tab->user_data, tab);
@@ -901,13 +945,13 @@ static gboolean multi_win_state_event_handler(GtkWidget *widget,
 
 static void multi_win_realize_handler(GtkWidget *win, gpointer data)
 {
-    GdkWindow *w = win->window;
+    GdkWindow *w = gtk_widget_get_window(win);
 
     gdk_window_set_group(w, w);
     (void) data;
 }
 
-static void multi_win_destroy_handler(GtkObject * obj, MultiWin * win)
+static void multi_win_destroy_handler(GObject * obj, MultiWin * win)
 {
     (void) obj;
 
@@ -931,7 +975,7 @@ static void multi_win_menutree_deleted_handler(MenuTree * tree, gpointer data)
 }
 
 static void
-multi_win_page_switched(GtkNotebook * notebook, GtkNotebookPage * page,
+multi_win_page_switched(GtkNotebook * notebook, GtkWidget *page,
     guint page_num, MultiWin * win)
 {
     MultiTab *tab = multi_tab_get_from_widget(
@@ -961,6 +1005,17 @@ static gboolean multi_win_focus_in(GtkWidget *widget, GdkEventFocus *event,
     return FALSE;
 }
 
+static void popup_tabs_menu(MultiWin *win, GdkEventButton *event)
+{
+    GtkMenuItem *mparent;
+    
+    g_return_if_fail(win->popup_menu != NULL);
+    mparent = GTK_MENU_ITEM(
+            menutree_get_widget_for_id(win->popup_menu, MENUTREE_TABS));
+    gtk_menu_popup(GTK_MENU(gtk_menu_item_get_submenu(mparent)),
+            NULL, NULL, NULL, NULL, event->button, event->time);
+}
+
 static gboolean tab_clicked_handler(GtkWidget *widget,
         GdkEventButton *event, MultiTab *tab)
 {
@@ -973,7 +1028,11 @@ static gboolean tab_clicked_handler(GtkWidget *widget,
             gtk_clipboard_request_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY),
                                        tab_set_name_from_clipboard_callback,
                                        tab);
-            break;
+            return TRUE;
+        case 3:
+            multi_win_select_tab(tab->parent, tab);
+            popup_tabs_menu(tab->parent, event);
+            return TRUE;
         default:
             break;
     }
@@ -1014,10 +1073,26 @@ static void page_added_callback(GtkNotebook *notebook, GtkWidget *child,
         }
         multi_win_add_tab(win, tab, page_num, TRUE);
         multi_tab_to_new_window_handler(win, tab, old_win_destroyed);
-        if (win->ntabs == 1)
+        if (win->tab_pos == GTK_POS_LEFT || win->tab_pos == GTK_POS_RIGHT)
         {
-            /* multi_tab_set_single_size(tab); */
-            multi_win_show(win);
+            multi_tab_pack_for_horizontal(tab, GTK_CONTAINER(notebook));
+        }
+        else
+        {
+            if (win->ntabs == 1)
+            {
+                multi_win_pack_for_single_tab(win);
+                /* multi_tab_set_single_size(tab); */
+                multi_win_show(win);
+            }
+            else if (win->ntabs == 2)
+            {
+                multi_win_pack_for_multiple_tabs(win);
+            }
+            else
+            {
+                multi_tab_pack_for_multiple(tab, GTK_CONTAINER(notebook));
+            }
         }
     }
 }
@@ -1053,7 +1128,8 @@ MultiWin *multi_win_new_for_tab(const char *display_name, int x, int y,
     multi_win_set_geometry_hints_for_tab(win, tab);
     gwin = GTK_WINDOW(win->gtkwin);
     gtk_window_set_default_size(gwin, w, h);
-    gtk_window_move(gwin, MAX(x - 20, 0), MAX(y - 8, 0));
+    if (x != -1 && y != -1)
+        gtk_window_move(gwin, MAX(x - 20, 0), MAX(y - 8, 0));
     gtk_widget_show_all(win->notebook);
     return win;
 }
@@ -1094,11 +1170,31 @@ static void multi_win_new_tab_action(MultiWin * win)
         win->user_data_template);
 }
 
+static void multi_win_detach_tab_action(MultiWin * win)
+{
+    MultiTab *tab = win->current_tab;
+    
+    win = multi_win_new_for_tab(win->display_name, -1, -1, tab);
+    multi_tab_move_to_new_window(win, tab, -1);
+    gtk_widget_show_all(win->gtkwin);
+}
+
 static void multi_win_close_tab_action(MultiWin * win)
 {
     multi_tab_delete(win->current_tab);
 }
 
+static void multi_win_close_other_tabs_action(MultiWin * win)
+{
+    GList *link;
+    
+    while (win->tabs->data != win->current_tab)
+        multi_tab_delete(win->tabs->data);
+    link = win->tabs;
+    while ((link = g_list_next(link)) != NULL)
+        multi_tab_delete(link->data);
+}
+
 static void multi_win_name_tab_action(MultiWin * win)
 {
     MultiTab *tab = win->current_tab;
@@ -1117,7 +1213,8 @@ static void multi_win_name_tab_action(MultiWin * win)
 
     tab->rename_dialog = dialog_w;
     gtk_dialog_set_default_response(dialog, GTK_RESPONSE_APPLY);
-    gtk_box_pack_start(GTK_BOX(dialog->vbox), name_w, TRUE, TRUE, 8);
+    gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(dialog)),
+            name_w, TRUE, TRUE, 8);
     gtk_entry_set_activates_default(name_e, TRUE);
     gtk_entry_set_text(name_e, name ? name : "");
     gtk_widget_show_all(dialog_w);
@@ -1169,13 +1266,14 @@ static void multi_win_set_window_title_action(MultiWin * win)
             "profile's title string."));
     int response;
     const char *title;
+    GtkBox *content_area = GTK_BOX(gtk_dialog_get_content_area(dialog));
 
     gtk_dialog_set_default_response(dialog, GTK_RESPONSE_APPLY);
-    gtk_box_pack_start(GTK_BOX(dialog->vbox), title_w, FALSE, TRUE, 8);
+    gtk_box_pack_start(content_area, title_w, FALSE, TRUE, 8);
     gtk_box_pack_start(GTK_BOX(hbox), img, FALSE, TRUE, 0);
     gtk_label_set_line_wrap(GTK_LABEL(tip_label), TRUE);
     gtk_box_pack_start(GTK_BOX(hbox), tip_label, TRUE, TRUE, 0);
-    gtk_box_pack_start(GTK_BOX(dialog->vbox), hbox, TRUE, TRUE, 8);
+    gtk_box_pack_start(content_area, hbox, TRUE, TRUE, 8);
     gtk_entry_set_activates_default(title_e, TRUE);
     gtk_entry_set_text(title_e, win->title_template ? win->title_template : "");
     gtk_widget_show_all(dialog_w);
@@ -1300,29 +1398,78 @@ static void multi_win_zoom_norm_action(MultiWin *win)
     }
 }
 
-static void multi_win_vscroll_by(MultiWin *win, double multiplier)
+typedef enum {
+    MULTI_TAB_SCROLL_STEP_LINE,
+    MULTI_TAB_SCROLL_STEP_PAGE,
+    MULTI_TAB_SCROLL_STEP_END
+} MultiTabScrollStep;
+
+static void multi_win_vscroll_by(MultiWin *win, double multiplier,
+        MultiTabScrollStep step_type)
 {
     GtkAdjustment *adj;
     double newval;
+    double bottom;
     
     adj = win->current_tab->adjustment;
     if (!adj)
         return;
-    newval = adj->value + multiplier * adj->step_increment;
-    newval = CLAMP(newval, 0, adj->upper - adj->page_size);
+    bottom = gtk_adjustment_get_upper(adj) - gtk_adjustment_get_page_size(adj);
+    if (step_type == MULTI_TAB_SCROLL_STEP_END)
+    {
+        if (multiplier < 0)
+            newval = 0;
+        else
+            newval = bottom;
+    }
+    else
+    {
+        double step;
+        
+        if (step_type == MULTI_TAB_SCROLL_STEP_PAGE)
+            step = gtk_adjustment_get_page_increment(adj);
+        else
+            step = gtk_adjustment_get_step_increment(adj);
+        newval = gtk_adjustment_get_value(adj) + multiplier * step;
+        newval = CLAMP(newval, 0, bottom);
+    }
     gtk_adjustment_set_value(adj, newval);
 }
 
 static void multi_win_scroll_up_action(MultiWin *win)
 {
     g_return_if_fail(win->current_tab != NULL);
-    multi_win_vscroll_by(win, -1);
+    multi_win_vscroll_by(win, -1, MULTI_TAB_SCROLL_STEP_LINE);
 }
 
 static void multi_win_scroll_down_action(MultiWin *win)
 {
     g_return_if_fail(win->current_tab != NULL);
-    multi_win_vscroll_by(win, 1);
+    multi_win_vscroll_by(win, 1, MULTI_TAB_SCROLL_STEP_LINE);
+}
+
+static void multi_win_scroll_page_up_action(MultiWin *win)
+{
+    g_return_if_fail(win->current_tab != NULL);
+    multi_win_vscroll_by(win, -1, MULTI_TAB_SCROLL_STEP_PAGE);
+}
+
+static void multi_win_scroll_page_down_action(MultiWin *win)
+{
+    g_return_if_fail(win->current_tab != NULL);
+    multi_win_vscroll_by(win, 1, MULTI_TAB_SCROLL_STEP_PAGE);
+}
+
+static void multi_win_scroll_to_top_action(MultiWin *win)
+{
+    g_return_if_fail(win->current_tab != NULL);
+    multi_win_vscroll_by(win, -1, MULTI_TAB_SCROLL_STEP_END);
+}
+
+static void multi_win_scroll_to_bottom_action(MultiWin *win)
+{
+    g_return_if_fail(win->current_tab != NULL);
+    multi_win_vscroll_by(win, 1, MULTI_TAB_SCROLL_STEP_END);
 }
 
 static void multi_win_move_tab_by_one(MultiWin *win, int dir)
@@ -1368,6 +1515,8 @@ static void multi_win_connect_actions(MultiWin * win)
         (multi_win_new_tab_action), win, NULL, NULL, NULL);
     multi_win_menu_connect_swapped(win, MENUTREE_FILE_CLOSE_TAB, G_CALLBACK
         (multi_win_close_tab_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_TABS_DETACH_TAB, G_CALLBACK
+        (multi_win_detach_tab_action), win, NULL, NULL, NULL);
     menutree_signal_connect_swapped(win->popup_menu,
             MENUTREE_FILE_NEW_WINDOW_WITH_PROFILE_HEADER,
             G_CALLBACK(multi_win_popup_new_term_with_profile),
@@ -1386,6 +1535,10 @@ static void multi_win_connect_actions(MultiWin * win)
         (multi_win_move_tab_right_action), win, NULL, NULL, NULL);
     multi_win_menu_connect_swapped(win, MENUTREE_TABS_NAME_TAB, G_CALLBACK
         (multi_win_name_tab_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_TABS_CLOSE_TAB, G_CALLBACK
+        (multi_win_close_tab_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_TABS_CLOSE_OTHER_TABS,
+        G_CALLBACK(multi_win_close_other_tabs_action), win, NULL, NULL, NULL);
     multi_win_menu_connect_swapped(win, MENUTREE_EDIT_SET_WINDOW_TITLE,
         G_CALLBACK(multi_win_set_window_title_action), win, NULL, NULL, NULL);
     multi_win_menu_connect(win, MENUTREE_VIEW_SHOW_TAB_BAR, G_CALLBACK
@@ -1400,10 +1553,18 @@ static void multi_win_connect_actions(MultiWin * win)
         (multi_win_zoom_out_action), win, NULL, NULL, NULL);
     multi_win_menu_connect_swapped(win, MENUTREE_VIEW_ZOOM_NORM, G_CALLBACK
         (multi_win_zoom_norm_action), win, NULL, NULL, NULL);
-    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_UP, G_CALLBACK
-        (multi_win_scroll_up_action), win, NULL, NULL, NULL);
-    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_DOWN, G_CALLBACK
-        (multi_win_scroll_down_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_UP,
+        G_CALLBACK(multi_win_scroll_up_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_DOWN,
+        G_CALLBACK(multi_win_scroll_down_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_PAGE_UP,
+        G_CALLBACK(multi_win_scroll_page_up_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_PAGE_DOWN,
+        G_CALLBACK(multi_win_scroll_page_down_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_TO_TOP,
+        G_CALLBACK(multi_win_scroll_to_top_action), win, NULL, NULL, NULL);
+    multi_win_menu_connect_swapped(win, MENUTREE_VIEW_SCROLL_TO_BOTTOM,
+        G_CALLBACK(multi_win_scroll_to_bottom_action), win, NULL, NULL, NULL);
 }
 
 static void multi_win_set_show_tabs_menu_items(MultiWin *win, gboolean active)
@@ -1432,10 +1593,30 @@ void multi_win_show(MultiWin *win)
     gtk_widget_show(win->notebook);
     gtk_widget_show(win->vbox);
     gtk_widget_show(win->gtkwin);
-    g_object_set_data(G_OBJECT(win->gtkwin->window), "ROXTermWin", win);
+    g_object_set_data(G_OBJECT(gtk_widget_get_window(win->gtkwin)),
+            "ROXTermWin", win);
 }
 
 #if HAVE_COMPOSITE
+#if GTK_CHECK_VERSION(3, 0, 0)
+void multi_win_set_colormap(MultiWin *win)
+{
+    GdkScreen *screen = gtk_widget_get_screen(win->gtkwin);
+    GdkVisual *visual = gdk_screen_get_rgba_visual(screen);
+
+    if (gdk_screen_is_composited(screen) && visual)
+    {
+        gtk_widget_set_visual(win->gtkwin, visual);
+        win->composite = TRUE;
+    }
+    else
+    {
+        gtk_widget_set_visual(win->gtkwin,
+                gdk_screen_get_system_visual(screen));
+        win->composite = FALSE;
+    }
+}
+#else
 void multi_win_set_colormap(MultiWin *win)
 {
     GdkScreen *screen = gtk_widget_get_screen(win->gtkwin);
@@ -1453,6 +1634,7 @@ void multi_win_set_colormap(MultiWin *win)
         win->composite = FALSE;
     }
 }
+#endif
 
 static void multi_win_composited_changed(GtkWidget *widget, MultiWin *win)
 {
@@ -1461,7 +1643,7 @@ static void multi_win_composited_changed(GtkWidget *widget, MultiWin *win)
     
     if (composited != win->composite)
     {
-        if (GTK_WIDGET_REALIZED(win->gtkwin))
+        if (gtk_widget_get_realized(win->gtkwin))
         {
             /* This section mostly copied from gnome-terminal */
             guint32 user_time;
@@ -1469,6 +1651,7 @@ static void multi_win_composited_changed(GtkWidget *widget, MultiWin *win)
             guint32 desktop = 0;
             gboolean was_minimized;
             int x, y;
+            GdkWindow *dwin = gtk_widget_get_window(win->gtkwin);
             
             user_time = gdk_x11_display_get_user_time(
                     gtk_widget_get_display (widget));
@@ -1476,8 +1659,8 @@ static void multi_win_composited_changed(GtkWidget *widget, MultiWin *win)
             /* If compositing changed, re-realize the window. Bug #563561 */
             
             gtk_window_get_position(GTK_WINDOW(win->gtkwin), &x, &y);
-            was_minimized = x11support_window_is_minimized(win->gtkwin->window);
-            have_desktop = x11support_get_wm_desktop(win->gtkwin->window,
+            was_minimized = x11support_window_is_minimized(dwin);
+            have_desktop = x11support_get_wm_desktop(dwin,
                     &desktop);
             gtk_widget_hide (widget);
             gtk_widget_unrealize (widget);
@@ -1486,14 +1669,14 @@ static void multi_win_composited_changed(GtkWidget *widget, MultiWin *win)
             
             gtk_window_move(GTK_WINDOW(win->gtkwin), x, y);
             gtk_widget_realize(win->gtkwin);
-            gdk_x11_window_set_user_time(win->gtkwin->window, user_time);
+            gdk_x11_window_set_user_time(dwin, user_time);
             if (was_minimized)
                 gtk_window_iconify(GTK_WINDOW(win->gtkwin));
             else
                 gtk_window_deiconify(GTK_WINDOW(win->gtkwin));
             gtk_widget_show(widget);
             if (have_desktop)
-                x11support_set_wm_desktop(win->gtkwin->window, desktop);
+                x11support_set_wm_desktop(dwin, desktop);
             win->clear_demands_attention = TRUE;
         }
         else
@@ -1508,7 +1691,7 @@ static gboolean multi_win_map_event_handler(GtkWidget *widget,
 {
     if (win->clear_demands_attention)
     {
-        x11support_clear_demands_attention(widget->window);
+        x11support_clear_demands_attention(gtk_widget_get_window(widget));
         win->clear_demands_attention = FALSE;
     }
     return FALSE;
@@ -1522,17 +1705,24 @@ MultiWin *multi_win_new_blank(const char *display_name, Options *shortcuts,
         GtkPositionType tab_pos, gboolean always_show_tabs)
 {
     MultiWin *win = g_new0(MultiWin, 1);
+#if !GTK_CHECK_VERSION(2, 24, 0)
     static gboolean set_nwc_hook = FALSE;
+#endif
     GtkNotebook *notebook;
     char *role = NULL;
 
+#if !GTK_CHECK_VERSION(2, 24, 0)
     if (!set_nwc_hook)
     {
         gtk_notebook_set_window_creation_hook(multi_win_notebook_creation_hook,
                 win, NULL);
         set_nwc_hook = TRUE;
     }
+#endif
 
+#if MULTITAB_LABEL_GTK3_SIZE_KLUDGE
+    win->best_tab_width = G_MAXINT;
+#endif
     win->tab_pos = tab_pos;
     win->always_show_tabs = always_show_tabs;
     win->scroll_bar_pos = MultiWinScrollBar_Query;
@@ -1575,7 +1765,7 @@ MultiWin *multi_win_new_blank(const char *display_name, Options *shortcuts,
             G_CALLBACK(multi_win_destroy_handler), win);
     if (multi_win_delete_handler)
     {
-        g_signal_connect(GTK_OBJECT(win->gtkwin), "delete-event",
+        g_signal_connect(win->gtkwin, "delete-event",
                 G_CALLBACK(multi_win_delete_handler), win);
     }
     
@@ -1627,7 +1817,12 @@ MultiWin *multi_win_new_blank(const char *display_name, Options *shortcuts,
 
     win->notebook = gtk_notebook_new();
     notebook = GTK_NOTEBOOK(win->notebook);
+#if GTK_CHECK_VERSION(2, 24, 0)
+    g_signal_connect(notebook, "create-window",
+            G_CALLBACK(multi_win_notebook_creation_hook), win);
+#endif
     gtk_notebook_set_scrollable(notebook, TRUE);
+#if !GTK_CHECK_VERSION(3, 0, 0)
     if (win->tab_pos == GTK_POS_LEFT || win->tab_pos == GTK_POS_RIGHT)
     {
         g_object_set(notebook, "tab-hborder", 0, NULL);
@@ -1636,7 +1831,8 @@ MultiWin *multi_win_new_blank(const char *display_name, Options *shortcuts,
     {
         g_object_set(notebook, "tab-vborder", 0, NULL);
     }
-    gtk_notebook_popup_enable(notebook);
+#endif
+    gtk_notebook_popup_disable(notebook);
     if (always_show_tabs)
     {
         multi_win_set_show_tabs_menu_items(win, TRUE);
@@ -1647,12 +1843,16 @@ MultiWin *multi_win_new_blank(const char *display_name, Options *shortcuts,
         multi_win_set_show_tabs_menu_items(win, FALSE);
         multi_win_hide_tabs(win);
     }
-    gtk_box_pack_start_defaults(GTK_BOX(win->vbox), win->notebook);
+    gtk_box_pack_start(GTK_BOX(win->vbox), win->notebook, TRUE, TRUE, 0);
     g_signal_connect(win->notebook, "switch-page",
         G_CALLBACK(multi_win_page_switched), win);
     g_signal_connect(win->gtkwin, "focus-in-event",
         G_CALLBACK(multi_win_focus_in), win);
+#if GTK_CHECK_VERSION(2, 24, 0)
+    gtk_notebook_set_group_name(notebook, "ROXTerm");
+#else
     gtk_notebook_set_group(notebook, &multi_win_all);
+#endif
     g_signal_connect(win->notebook, "page-reordered",
         G_CALLBACK(page_reordered_callback), win);
     g_signal_connect(win->notebook, "page-added",
@@ -1706,6 +1906,8 @@ MultiWin *multi_win_new_full(const char *display_name, Options *shortcuts,
 
     if (geom)
     {
+        /* Need to show children before parsing geom */
+        gtk_widget_show_all(win->vbox);
         gtk_window_parse_geometry(GTK_WINDOW(win->gtkwin), geom);
     }
     else if (sizing == MULTI_WIN_FULL_SCREEN)
@@ -1809,6 +2011,8 @@ static gboolean multi_win_notify_tab_removed(MultiWin * win, MultiTab * tab)
         if (win->ntabs == 1)
         {
             tab = win->tabs->data;
+            if (win->tab_pos == GTK_POS_TOP || win->tab_pos == GTK_POS_BOTTOM)
+                multi_win_pack_for_single_tab(win);
             /*multi_tab_set_single_size(tab);*/
             if (!win->always_show_tabs)
             {
@@ -1823,34 +2027,18 @@ static gboolean multi_win_notify_tab_removed(MultiWin * win, MultiTab * tab)
 
 void multi_tab_add_close_button(MultiTab *tab)
 {
-    GtkRcStyle *rcstyle;
-    GtkSettings *settings;
-    int w, h;
     MultiWin *win;
 
     if (tab->close_button)
         return;
     
     win = tab->parent;
-    tab->close_button = gtk_button_new();
-    tab->image = gtk_image_new_from_stock(
-            tab->status_stock ? tab->status_stock : GTK_STOCK_CLOSE,
-            GTK_ICON_SIZE_MENU);
-    rcstyle = gtk_rc_style_new();
-    gtk_container_add(GTK_CONTAINER(tab->close_button), tab->image);
-    gtk_button_set_relief(GTK_BUTTON(tab->close_button), GTK_RELIEF_NONE);
-    gtk_button_set_focus_on_click(GTK_BUTTON(tab->close_button), FALSE);
-    settings = gtk_widget_get_settings(tab->label);
-    gtk_icon_size_lookup_for_settings(settings, GTK_ICON_SIZE_MENU, &w, &h);
-    /* gtk_widget_set_size_request(tab->close_button, w + 2, h + 2); */
-    rcstyle->xthickness = rcstyle->ythickness = 0;
-    gtk_widget_modify_style(tab->close_button, rcstyle);
-    gtk_rc_style_unref(rcstyle);
-    tab->label_packer(GTK_BOX(tab->label_box), tab->close_button,
+    tab->close_button = multitab_close_button_new(tab->status_stock);
+    gtk_box_pack_start(GTK_BOX(tab->label_box), tab->close_button,
             FALSE, FALSE, 0);
     g_signal_connect(tab->close_button, "clicked",
             G_CALLBACK(multi_win_close_tab_clicked), tab);
-    gtk_widget_show_all(tab->close_button);
+    gtk_widget_show(tab->close_button);
     if (win)
     {
         multi_win_restore_size(win);
@@ -1871,7 +2059,6 @@ void multi_tab_remove_close_button(MultiTab *tab)
             multi_win_restore_size(win);
         }
     }
-    tab->image = NULL;
 }
 
 void multi_tab_set_status_stock(MultiTab *tab, const char *stock)
@@ -1882,10 +2069,10 @@ void multi_tab_set_status_stock(MultiTab *tab, const char *stock)
         return;
     }
     tab->status_stock = stock;
-    if (tab->image)
+    if (tab->close_button)
     {
-        gtk_image_set_from_stock(GTK_IMAGE(tab->image),
-                stock ? stock : GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
+        multitab_close_button_set_image(
+                MULTITAB_CLOSE_BUTTON(tab->close_button), stock);
     }
 }
 
@@ -1893,42 +2080,16 @@ void multi_tab_set_status_stock(MultiTab *tab, const char *stock)
  * the text; the return value is the top-level container. */
 static GtkWidget *make_tab_label(MultiTab *tab, GtkPositionType tab_pos)
 {
-    static GdkColor amber;
-    static gboolean parsed_amber = FALSE;
-    GtkLabel *label;
-
-    tab->label = gtk_label_new(NULL);
-    tab->menu_label = gtk_label_new(NULL);
-    label = GTK_LABEL(tab->label);
-    gtk_label_set_ellipsize(label, PANGO_ELLIPSIZE_MIDDLE);
+#if MULTITAB_LABEL_GTK3_SIZE_KLUDGE
+    tab->label = multitab_label_new(NULL, &tab->parent->best_tab_width);
+#else
+    tab->label = multitab_label_new(NULL);
+#endif
     multi_tab_set_full_window_title(tab, tab->window_title_template,
             tab->window_title);
-    switch (tab_pos)
-    {
-        case GTK_POS_LEFT:
-        case GTK_POS_RIGHT:
-            tab->label_box = gtk_vbox_new(FALSE, 4);
-            tab->label_packer = gtk_box_pack_end;
-            gtk_widget_set_size_request(tab->label, HORIZ_TAB_WIDTH, -1);
-            break;
-        default:
-            tab->label_box = gtk_hbox_new(FALSE, 4);
-            tab->label_packer = gtk_box_pack_start;
-            break;
-    }
-    
-    tab->label_bg = gtk_event_box_new();
-    if (!parsed_amber)
-    {
-        gdk_color_parse("#ffc450", &amber);
-        parsed_amber = TRUE;
-    }
-    gtk_widget_modify_bg(tab->label_bg, GTK_STATE_NORMAL, &amber);
-    gtk_widget_modify_bg(tab->label_bg, GTK_STATE_ACTIVE, &amber);
-    gtk_event_box_set_visible_window(GTK_EVENT_BOX(tab->label_bg), FALSE);
-    tab->label_packer(GTK_BOX(tab->label_box), tab->label_bg, TRUE, TRUE, 0);
-    gtk_container_add(GTK_CONTAINER(tab->label_bg), tab->label);
-    g_signal_connect(tab->label_bg, "button-press-event",
+    tab->label_box = gtk_hbox_new(FALSE, 4);
+    gtk_box_pack_start(GTK_BOX(tab->label_box), tab->label, TRUE, TRUE, 0);
+    g_signal_connect(tab->label, "button-press-event",
             G_CALLBACK(tab_clicked_handler), tab);
     if (multi_tab_get_show_close_button(tab->user_data))
     {
@@ -1970,13 +2131,11 @@ static void multi_win_add_tab_to_notebook(MultiWin * win, MultiTab * tab,
 
     if (position == -1)
     {
-        gtk_notebook_append_page_menu(notebook, tab->widget, label,
-                tab->menu_label);
+        gtk_notebook_append_page(notebook, tab->widget, label);
     }
     else
     {
-        gtk_notebook_insert_page_menu(notebook,
-                tab->widget, label, tab->menu_label, position);
+        gtk_notebook_insert_page(notebook, tab->widget, label, position);
     }
     gtk_notebook_set_tab_reorderable(notebook, tab->widget, TRUE);
     gtk_notebook_set_tab_detachable(notebook, tab->widget, TRUE);
@@ -1984,27 +2143,19 @@ static void multi_win_add_tab_to_notebook(MultiWin * win, MultiTab * tab,
     /* Note at this point ntabs is how many tabs there are about to be,
      * not how many there were before adding.
      */
-    /*
-    if (win->ntabs == 1)
+    if (win->tab_pos == GTK_POS_TOP || win->tab_pos == GTK_POS_BOTTOM)
     {
-        multi_tab_set_single_size(tab);
+        if (win->ntabs == 1)
+            multi_win_pack_for_single_tab(win);
+        else if (win->ntabs == 2)
+            multi_win_pack_for_multiple_tabs(win);
+        else
+            multi_tab_pack_for_multiple(tab, GTK_CONTAINER(notebook));
     }
     else
     {
-        GList *link;
-        
-        g_object_set(G_OBJECT(win->notebook), "homogeneous", TRUE, NULL);
-        for (link = win->tabs; link; link = g_list_next(link))
-        {
-            MultiTab *tab = link->data;
-            
-            gtk_container_child_set(GTK_CONTAINER(notebook), tab->widget,
-                    "tab-expand", TRUE, "tab-fill", TRUE, NULL);
-        }
+        multi_tab_pack_for_horizontal(tab, GTK_CONTAINER(notebook));
     }
-    */
-    gtk_container_child_set(GTK_CONTAINER(notebook), tab->widget,
-            "tab-expand", TRUE, "tab-fill", TRUE, NULL);
 }
 
 /* Keep track of a tab after it's been created; if notify_only is set, it's
@@ -2030,6 +2181,8 @@ static void multi_win_add_tab(MultiWin * win, MultiTab * tab, int position,
     {
         multi_win_add_tab_to_notebook(win, tab, position);
         gtk_widget_show(tab->widget);
+        if (win->ntabs > 1)
+            gtk_window_present(GTK_WINDOW(win->gtkwin));
     }
     multi_win_shade_menus_for_tabs(win);
     multi_win_select_tab(win, tab);
@@ -2048,9 +2201,9 @@ gboolean multi_win_is_fullscreen(MultiWin *win)
 
 gboolean multi_win_is_maximised(MultiWin *win)
 {
-    return (win->gtkwin->window &&
-            (gdk_window_get_state(win->gtkwin->window) &
-                GDK_WINDOW_STATE_MAXIMIZED) != 0);
+    GdkWindow *w = gtk_widget_get_window(win->gtkwin);
+    
+    return (w && (gdk_window_get_state(w) & GDK_WINDOW_STATE_MAXIMIZED) != 0);
 }
 
 int multi_win_get_zoom_index(MultiWin *win)
@@ -2138,16 +2291,20 @@ Options *multi_win_get_shortcut_scheme(MultiWin * win)
     return win->shortcuts;
 }
 
-void multi_win_set_shortcut_scheme(MultiWin *win, Options *shortcuts)
+void multi_win_set_shortcut_scheme(MultiWin *win, Options *shortcuts,
+        gboolean reapply)
 {
-    if (win->shortcuts != shortcuts)
+    if (win->shortcuts != shortcuts || reapply)
     {
-        if (win->shortcuts)
+        if (win->shortcuts && win->shortcuts != shortcuts)
         {
             UNREF_LOG(shortcuts_unref(win->shortcuts));
         }
-        options_ref(shortcuts);
-        win->shortcuts = shortcuts;
+        if (win->shortcuts != shortcuts)
+        {
+            options_ref(shortcuts);
+            win->shortcuts = shortcuts;
+        }
         if (win->menu_bar)
             menutree_apply_shortcuts(win->menu_bar, shortcuts);
         if (win->popup_menu)
