@@ -1,8 +1,8 @@
 #ifndef MULTITAB_H
 #define MULTITAB_H
 /*
-    roxterm - GTK+ 2.0 terminal emulator with tabs
-    Copyright (C) 2004 Tony Houghton <h@realh.co.uk>
+    roxterm - VTE/GTK terminal emulator with tabs
+    Copyright (C) 2004-2011 Tony Houghton <h@realh.co.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -102,6 +102,9 @@ typedef gboolean (*MultiTabGetShowCloseButton)(gpointer user_data);
 /* Needed externally */
 MultiWinGetDisableMenuShortcuts multi_win_get_disable_menu_shortcuts;
 
+/* Whether to open a new tab adjacent to current one */
+typedef gboolean (*MultiTabGetNewTabAdjacent)(gpointer user_data);
+
 /* Call to set up function hooks. See MultiTabFiller etc above.
  * menu_signal_connector is called each time a new window is created to give
  * the client a chance to connect its signal handlers; each handler will
@@ -112,7 +115,8 @@ multi_tab_init(MultiTabFiller filler, MultiTabDestructor destructor,
     MultiWinGeometryFunc, MultiWinSizeFunc, MultiWinDefaultSizeFunc,
     MultiTabToNewWindowHandler,
     MultiWinZoomHandler, MultiWinGetDisableMenuShortcuts, MultiWinInitialTabs,
-    MultiWinDeleteHandler, MultiTabGetShowCloseButton
+    MultiWinDeleteHandler, MultiTabGetShowCloseButton,
+    MultiTabGetNewTabAdjacent
     );
 
 /* Register a MultiTabSelectionHandler (see above) */
@@ -343,7 +347,8 @@ MenuTree *multi_win_get_short_popup_menu(MultiWin * win);
 
 Options *multi_win_get_shortcut_scheme(MultiWin * win);
 
-void multi_win_set_shortcut_scheme(MultiWin * win, Options *);
+/* reapply is for forcing an update of the current scheme */
+void multi_win_set_shortcut_scheme(MultiWin * win, Options *, gboolean reapply);
 
 GtkAccelGroup *multi_win_get_accel_group(MultiWin * win);
 
