@@ -1,6 +1,6 @@
 /*
-    roxterm - GTK+ 2.0 terminal emulator with tabs
-    Copyright (C) 2004 Tony Houghton <h@realh.co.uk>
+    roxterm - VTE/GTK terminal emulator with tabs
+    Copyright (C) 2004-2011 Tony Houghton <h@realh.co.uk>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,10 +20,9 @@
 
 #include "defns.h"
 
+#include <stdarg.h>
 #include <string.h>
 
-#include <glade/glade-xml.h>
-
 #include "capplet.h"
 #include "configlet.h"
 #include "dlg.h"
@@ -34,127 +33,89 @@
 
 #define MANAGE_PREVIEW
 
-#define PG_GLADE_CONNECT(pg, glade, h) \
-    glade_xml_signal_connect_data ((glade), #h, (GCallback) (h), (pg))
-
-#define PG_CONNECT(pg, h) PG_GLADE_CONNECT((pg), (pg)->glade, h)
-
-#define PG_GLADE_CONNECT_OPTS(pg, glade, h) \
-    glade_xml_signal_connect_data ((glade), #h, (GCallback) (h), \
-            (pg)->profile)
-
-#define PG_CONNECT_OPTS(pg, h) PG_GLADE_CONNECT_OPTS((pg), (pg)->glade, h)
-
-#define PG_GLADE_EC_CONNECT(pg, glade, n) \
-    glade_xml_signal_connect_data ((glade), "on_editable_changed", \
-            (GCallback) on_editable_changed, &pg->changed. n)
-
-#define PG_EC_CONNECT(pg, n) PG_GLADE_EC_CONNECT((pg), (pg)->glade, n)
-
-struct ProfileGUI {
-    GladeXML *glade;
-    GladeXML *ssh_glade;
-    GtkWidget *widget;
+struct _ProfileGUI {
+    CappletData capp;
+    GtkWidget *dialog;
+    GtkWidget *ssh_dialog;
     char *profile_name;
-    Options *profile;
     gboolean ignore_destroy;
     struct {
-        char sel_by_word;
-        char color_term;
-        char term;
-        char browser;
-        char mailer;
-        char command;
-        char title_string;
-        char ssh_address;
-        char ssh_user;
-        char ssh_options;
-        char win_title;
+        guint sel_by_word : 1;
+        guint color_term : 1;
+        guint term : 1;
+        guint browser : 1;
+        guint mailer : 1;
+        guint filer : 1;
+        guint dir_filer : 1;
+        guint command : 1;
+        guint title_string : 1;
+        guint ssh_address : 1;
+        guint ssh_user : 1;
+        guint ssh_options : 1;
+        guint win_title : 1;
     } changed;
-    GtkWidget *bgimg;
-#ifndef HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE
-    GtkWidget *cursor_blinks;
-#endif
     DragReceiveData *bgimg_drd;
     GtkListStore *list_store;
 };
 
 static GHashTable *profilegui_being_edited = NULL;
 
+#if 0
 inline static
 GtkWidget *profilegui_widget(ProfileGUI *pg, const char *name)
 {
-    return glade_xml_get_widget(pg->glade, name);
+    /* Result will often be cast to a specialised widget so use static
+     * cast here to avoid double GObject cast.
+     */
+    return (GtkWidget *) gtk_builder_get_object(pg->capp.builder, name);
+}
+#endif
+static
+GtkWidget *profilegui_widget(ProfileGUI *pg, const char *name)
+{
+    GtkWidget *widget =
+            (GtkWidget *)gtk_builder_get_object(pg->capp.builder, name);
+    
+    if (!widget)
+        g_critical(_("Profile widget '%s' not found"), name);
+    return widget;
 }
 
-static void update_from_entry(ProfileGUI * pg, GladeXML *glade,
-        const char *name)
+static void profilegui_update_from_entry(ProfileGUI * pg, const char *name)
 {
-    capplet_set_string(pg->profile, name, gtk_entry_get_text
-        (GTK_ENTRY(glade_xml_get_widget(glade, name))));
+    capplet_set_string(pg->capp.options, name,
+            gtk_entry_get_text(
+                    GTK_ENTRY(gtk_builder_get_object(pg->capp.builder, name))));
 }
 
-inline static void
-profilegui_update_from_entry(ProfileGUI *pg, const char *name)
-{
-    update_from_entry(pg, pg->glade, name);
-}
+#define PG_MEMBER(a, b) a . b
+
+#define PG_UPDATE_IF(s) if (PG_MEMBER(pg->changed, s)) \
+    { \
+        profilegui_update_from_entry(pg, # s); \
+        PG_MEMBER(pg->changed, s) = 0; \
+    }
 
 void profilegui_check_entries_for_changes(ProfileGUI * pg)
 {
-    if (pg->changed.sel_by_word)
-    {
-        profilegui_update_from_entry(pg, "sel_by_word");
-        pg->changed.sel_by_word = 0;
-    }
-    if (pg->changed.color_term)
-    {
-        profilegui_update_from_entry(pg, "color_term");
-        pg->changed.color_term = 0;
-    }
-    if (pg->changed.term)
-    {
-        profilegui_update_from_entry(pg, "term");
-        pg->changed.term = 0;
-    }
-    if (pg->changed.browser)
-    {
-        profilegui_update_from_entry(pg, "browser");
-        pg->changed.browser = 0;
-    }
-    if (pg->changed.mailer)
-    {
-        profilegui_update_from_entry(pg, "mailer");
-        pg->changed.mailer = 0;
-    }
-    if (pg->changed.command)
-    {
-        profilegui_update_from_entry(pg, "command");
-        pg->changed.command = 0;
-    }
-    if (pg->changed.title_string)
-    {
-        profilegui_update_from_entry(pg, "title_string");
-        pg->changed.title_string = 0;
-    }
-    if (pg->changed.win_title)
-    {
-        profilegui_update_from_entry(pg, "win_title");
-        pg->changed.win_title = 0;
-    }
+    PG_UPDATE_IF(sel_by_word)
+    PG_UPDATE_IF(color_term)
+    PG_UPDATE_IF(term)
+    PG_UPDATE_IF(browser)
+    PG_UPDATE_IF(mailer)
+    PG_UPDATE_IF(filer)
+    PG_UPDATE_IF(dir_filer)
+    PG_UPDATE_IF(command)
+    PG_UPDATE_IF(title_string)
+    PG_UPDATE_IF(win_title)
 }
 
 static void profilegui_set_bgimg_shading(ProfileGUI *pg, gboolean sensitive)
 {
     gtk_widget_set_sensitive(profilegui_widget(pg,
                 "scroll_background"), sensitive);
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
-    gtk_widget_set_sensitive(pg->bgimg, sensitive);
-#else
-    gtk_widget_set_sensitive(profilegui_widget(pg, "bgimg_browse"), sensitive);
     gtk_widget_set_sensitive(profilegui_widget(pg, "background_img"),
             sensitive);
-#endif
     gtk_widget_set_sensitive(profilegui_widget(pg, "img_label"),
             sensitive);
 }
@@ -199,70 +160,59 @@ static void profilegui_set_close_buttons_shading(ProfileGUI *pg)
 
 static void profilegui_check_ssh_entries_for_changes(ProfileGUI * pg)
 {
-    GladeXML *glade = pg->ssh_glade;
-    
-    if (pg->changed.ssh_address)
-    {
-        update_from_entry(pg, glade, "ssh_address");
-        pg->changed.ssh_address = 0;
-    }
-    if (pg->changed.ssh_user)
-    {
-        update_from_entry(pg, glade, "ssh_user");
-        pg->changed.ssh_user = 0;
-    }
-    if (pg->changed.ssh_options)
-    {
-        update_from_entry(pg, glade, "ssh_options");
-        pg->changed.ssh_options = 0;
-    }
+    PG_UPDATE_IF(ssh_address)
+    PG_UPDATE_IF(ssh_user)
+    PG_UPDATE_IF(ssh_options)
 }
 
 /***********************************************************/
 /* Handlers */
 
-static void on_ssh_entry_activate(GtkEntry * entry, ProfileGUI * pg)
+void on_editable_changed(GtkEditable * editable, ProfileGUI *pg)
+{
+    const char *n = gtk_buildable_get_name(GTK_BUILDABLE(editable));
+#define PG_IF_CHANGED(s) if (!strcmp(n, #s)) PG_MEMBER(pg->changed, s) = 1
+
+    PG_IF_CHANGED(sel_by_word);
+    else PG_IF_CHANGED(color_term);
+    else PG_IF_CHANGED(term);
+    else PG_IF_CHANGED(browser);
+    else PG_IF_CHANGED(mailer);
+    else PG_IF_CHANGED(dir_filer);
+    else PG_IF_CHANGED(command);
+    else PG_IF_CHANGED(title_string);
+    else PG_IF_CHANGED(ssh_address);
+    else PG_IF_CHANGED(ssh_user);
+    else PG_IF_CHANGED(ssh_options);
+    else PG_IF_CHANGED(win_title);
+}
+
+void on_ssh_entry_activate(GtkEntry * entry, ProfileGUI * pg)
 {
     profilegui_check_ssh_entries_for_changes(pg);
 }
 
-static void on_ssh_host_changed(GtkWidget * widget, ProfileGUI *pg)
+void on_ssh_host_changed(GtkWidget * widget, ProfileGUI *pg)
 {
     gtk_entry_set_text(GTK_ENTRY(profilegui_widget(pg, "ssh_host")),
             gtk_entry_get_text(GTK_ENTRY(widget)));
 }
 
-static void on_edit_ssh_clicked(GtkWidget * widget, ProfileGUI *pg)
+void on_edit_ssh_clicked(GtkWidget * widget, ProfileGUI *pg)
 {
-    GladeXML *glade = glade_xml_new(capplet_get_glade_filename(),
-        "ssh_dialog", NULL);
-    GtkWidget *dialog = glade_xml_get_widget(glade, "ssh_dialog");
-    
-    pg->ssh_glade = glade;
     capplet_inc_windows();
     
-    PG_GLADE_EC_CONNECT(pg, glade, ssh_address);
-    PG_GLADE_EC_CONNECT(pg, glade, ssh_user);
-    PG_GLADE_EC_CONNECT(pg, glade, ssh_options);
-    PG_GLADE_CONNECT(pg, glade, on_ssh_entry_activate);
-    PG_GLADE_CONNECT(pg, glade, on_ssh_host_changed);
-    PG_GLADE_CONNECT_OPTS(pg, glade, on_spin_button_changed);
-
-    capplet_set_text_entry(glade, pg->profile, "ssh_address", "localhost");
-    capplet_set_text_entry(glade, pg->profile, "ssh_user", NULL);
-    capplet_set_spin_button(glade, pg->profile, "ssh_port", 22);
-    capplet_set_text_entry(glade, pg->profile, "ssh_options", NULL);
-    
-    gtk_dialog_run(GTK_DIALOG(dialog));
-    
+    capplet_set_text_entry(&pg->capp, "ssh_address", "localhost");
+    capplet_set_text_entry(&pg->capp, "ssh_user", NULL);
+    capplet_set_spin_button(&pg->capp, "ssh_port", 22);
+    capplet_set_text_entry(&pg->capp, "ssh_options", NULL);
+    gtk_dialog_run(GTK_DIALOG(pg->ssh_dialog));
     profilegui_check_ssh_entries_for_changes(pg);
-    gtk_widget_destroy(dialog);
-    UNREF_LOG(g_object_unref(glade));
-    pg->ssh_glade = NULL;
+    gtk_widget_hide(pg->ssh_dialog);
     capplet_dec_windows();
 }
 
-static void on_Profile_Editor_destroy(GtkWidget * widget, ProfileGUI * pg)
+void on_Profile_Editor_destroy(GtkWidget * widget, ProfileGUI * pg)
 {
     if (pg->ignore_destroy)
     {
@@ -275,54 +225,51 @@ static void on_Profile_Editor_destroy(GtkWidget * widget, ProfileGUI * pg)
     }
 }
 
-static void on_Profile_Editor_response(GtkWidget * widget, int response,
-        ProfileGUI * pg)
+void on_Profile_Editor_response(GtkWidget *widget, int response, ProfileGUI *pg)
 {
     profilegui_delete(pg);
 }
 
-static void on_Profile_Editor_close(GtkWidget * widget, ProfileGUI * pg)
+void on_Profile_Editor_close(GtkWidget * widget, ProfileGUI * pg)
 {
     profilegui_delete(pg);
 }
 
-static void on_font_set(GtkFontButton * fontbutton, ProfileGUI * pg)
+void on_font_set(GtkFontButton * fontbutton, ProfileGUI * pg)
 {
-    capplet_set_string(pg->profile,
+    capplet_set_string(pg->capp.options,
         "font", gtk_font_button_get_font_name(fontbutton));
 }
 
-static void on_entry_activate(GtkEntry * entry, ProfileGUI * pg)
+void on_entry_activate(GtkEntry * entry, ProfileGUI * pg)
 {
     profilegui_check_entries_for_changes(pg);
 }
 
-static void
-on_profile_notebook_switch_page(GtkNotebook * notebook, GtkNotebookPage * page,
-    guint page_num, ProfileGUI * pg)
+void on_profile_notebook_switch_page(GtkNotebook * notebook, GtkWidget *page,
+        guint page_num, ProfileGUI * pg)
 {
     profilegui_check_entries_for_changes(pg);
 }
 
-static void on_bgtype_toggled(GtkToggleButton *button, ProfileGUI *pg)
+void on_bgtype_toggled(GtkToggleButton *button, ProfileGUI *pg)
 {
     if (gtk_toggle_button_get_active(button))
         profilegui_set_background_shading(pg);
-    on_radio_toggled(button, pg->profile);
+    on_radio_toggled(button, &pg->capp);
 }
 
-static void on_command_toggled(GtkToggleButton *button, ProfileGUI *pg)
+void on_command_toggled(GtkToggleButton *button, ProfileGUI *pg)
 {
     profilegui_set_command_shading(pg);
 }
 
-static void on_close_buttons_toggled(GtkToggleButton *button, ProfileGUI *pg)
+void on_close_buttons_toggled(GtkToggleButton *button, ProfileGUI *pg)
 {
     profilegui_set_close_buttons_shading(pg);
 }
 
-#ifndef HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE
-static void on_cursor_blinks_toggled(GtkToggleButton *button, Options *opts)
+void on_cursor_blinks_toggled(GtkToggleButton *button, ProfileGUI *pg)
 {
     gboolean state;
     
@@ -330,9 +277,16 @@ static void on_cursor_blinks_toggled(GtkToggleButton *button, Options *opts)
         return;
     
     state = gtk_toggle_button_get_active(button);
-    capplet_set_int(opts, "cursor_blinks", state);
+    capplet_set_int(pg->capp.options, "cursor_blinks", state);
+}
+
+void on_cell_size_toggled(GtkToggleButton *button, ProfileGUI *pg)
+{
+    gboolean state = gtk_toggle_button_get_active(button);
+    
+    gtk_widget_set_sensitive(profilegui_widget(pg, "width"), state);
+    gtk_widget_set_sensitive(profilegui_widget(pg, "height"), state);
 }
-#endif
 
 #define PREVIEW_SIZE 160
 
@@ -370,6 +324,9 @@ pg_load_preview(GtkFileChooser *chooser, GtkWidget *preview,
                 PREVIEW_SIZE, PREVIEW_SIZE, TRUE, &error);
     }
     gtk_image_set_from_pixbuf(GTK_IMAGE(preview), pixbuf);
+    
+    /* Let's not report file not found any more */
+#if 0
     if (!pixbuf && filename && filename[0] &&
             !g_file_test(filename, G_FILE_TEST_IS_DIR))
     {
@@ -388,6 +345,8 @@ pg_load_preview(GtkFileChooser *chooser, GtkWidget *preview,
                     filename, error ? error->message : _("unknown reason"));
         }
     }
+#endif
+
     if (error)
         g_error_free(error);
     gtk_file_chooser_set_preview_widget_active(chooser, pixbuf != NULL);
@@ -410,9 +369,12 @@ static void on_update_preview(GtkFileChooser *chooser, ProfileGUI *pg)
 }
 #endif
 
-static void profilegui_setup_file_chooser(ProfileGUI *pg,
-        GtkFileChooser *chooser)
+static void profilegui_setup_file_chooser(ProfileGUI *pg)
 {
+    GtkFileChooser *chooser =
+            GTK_FILE_CHOOSER(gtk_builder_get_object(pg->capp.builder,
+                    "background_img"));
+                    
     GtkFileFilter *img_filter = gtk_file_filter_new();
     GtkFileFilter *all_filter = gtk_file_filter_new();
 #ifdef MANAGE_PREVIEW
@@ -435,8 +397,12 @@ static void profilegui_setup_file_chooser(ProfileGUI *pg,
 }
 
 static void profilegui_fill_in_file_chooser(ProfileGUI *pg,
-        GtkFileChooser *chooser, const char *old_filename)
+        const char *old_filename)
 {
+    GtkFileChooser *chooser =
+            GTK_FILE_CHOOSER(gtk_builder_get_object(pg->capp.builder,
+                    "background_img"));
+                    
 #ifdef MANAGE_PREVIEW
     GtkWidget *preview = gtk_file_chooser_get_preview_widget(chooser);
 #endif
@@ -456,61 +422,27 @@ static char *get_bgimg_filename(ProfileGUI *pg, GtkFileChooser *chooser)
         g_free(new_filename);
         return NULL;
     }
-    capplet_set_string(pg->profile, "background_img", new_filename);
+    capplet_set_string(pg->capp.options, "background_img", new_filename);
     return new_filename;
 }
 
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
-static void on_bgimg_chosen(GtkFileChooser *chooser, ProfileGUI *pg)
+void on_bgimg_chosen(GtkFileChooser *chooser, ProfileGUI *pg)
 {
     g_free(get_bgimg_filename(pg, chooser));
 }
-#else
-static void on_bgimg_browse_clicked(GtkButton *button, ProfileGUI *pg)
-{
-    GtkFileChooser *chooser =
-        GTK_FILE_CHOOSER(gtk_file_chooser_dialog_new(_("Background image"),
-        GTK_WINDOW(pg->widget),
-        GTK_FILE_CHOOSER_ACTION_OPEN,
-        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-        _("Use as background"), GTK_RESPONSE_ACCEPT,
-        NULL));
-    GtkEntry *bgimg_display = GTK_ENTRY(profilegui_widget(pg,
-                "background_img"));
-    const char *old_filename = gtk_entry_get_text(bgimg_display);
-    
-    profilegui_setup_file_chooser(pg, chooser);
-    profilegui_fill_in_file_chooser(pg, chooser, old_filename);
-    if (gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT)
-    {
-        char *new_filename = get_bgimg_filename(pg, chooser);
-
-        if (new_filename)
-        {
-            gtk_entry_set_text(bgimg_display, new_filename);
-            g_free(new_filename);
-        }
-    }
-    /*
-    UNREF_LOG(g_object_unref(img_filter));
-    UNREF_LOG(g_object_unref(all_filter));
-    */
-    gtk_widget_destroy(GTK_WIDGET(chooser));
-}
-#endif /* HAVE_GTK_FILE_CHOOSER_BUTTON_NEW */
 
 #define DEFAULT_BACKSPACE_BINDING 0
 #define DEFAULT_DELETE_BINDING 0
 
-static void on_reset_compat_clicked(GtkButton *button, ProfileGUI *pg)
+void on_reset_compat_clicked(GtkButton *button, ProfileGUI *pg)
 {
-    capplet_set_int(pg->profile, "backspace_binding",
+    capplet_set_int(pg->capp.options, "backspace_binding",
             DEFAULT_BACKSPACE_BINDING);
-    capplet_set_combo(pg->glade, pg->profile, "backspace_binding",
+    capplet_set_combo(&pg->capp, "backspace_binding",
             DEFAULT_BACKSPACE_BINDING);
-    capplet_set_int(pg->profile, "delete_binding",
+    capplet_set_int(pg->capp.options, "delete_binding",
             DEFAULT_DELETE_BINDING);
-    capplet_set_combo(pg->glade, pg->profile, "delete_binding",
+    capplet_set_combo(&pg->capp, "delete_binding",
             DEFAULT_DELETE_BINDING);
 }
 
@@ -534,13 +466,12 @@ static gboolean bgimg_drag_data_received(GtkWidget *widget,
 
     if (gdk_pixbuf_get_file_info(filename, NULL, NULL))
     {
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
-        gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(pg->bgimg), filename);
+        GtkFileChooser *chooser =
+                GTK_FILE_CHOOSER(gtk_builder_get_object(pg->capp.builder,
+                        "background_img"));
+                    
+        gtk_file_chooser_set_filename(chooser, filename);
         /* Don't need to call capplet_set_string because we'll get a signal */
-#else
-        gtk_entry_set_text(GTK_ENTRY(pg->bgimg), filename);
-        capplet_set_string(pg->profile, "background_image", filename);
-#endif
         gtk_toggle_button_set_active(
                 GTK_TOGGLE_BUTTON(profilegui_widget(pg, "background_type1")),
                 TRUE);
@@ -561,55 +492,12 @@ static void exit_action_changed(GtkComboBox *combo, ProfileGUI *pg)
 
 static void profilegui_connect_handlers(ProfileGUI * pg)
 {
-    /* Generic widgets */
-    PG_CONNECT_OPTS(pg, on_boolean_toggled);
-    PG_CONNECT_OPTS(pg, on_radio_toggled);
-    PG_CONNECT_OPTS(pg, on_spin_button_changed);
-    PG_CONNECT_OPTS(pg, on_float_spin_button_changed);
-    PG_CONNECT_OPTS(pg, on_float_range_changed);
-    PG_CONNECT_OPTS(pg, on_combo_changed);
-    PG_CONNECT(pg, on_entry_activate);
-
-    /* Editables */
-    PG_EC_CONNECT(pg, sel_by_word);
-    PG_EC_CONNECT(pg, color_term);
-    PG_EC_CONNECT(pg, term);
-    PG_EC_CONNECT(pg, command);
-    PG_EC_CONNECT(pg, browser);
-    PG_EC_CONNECT(pg, mailer);
-    PG_EC_CONNECT(pg, title_string);
-    PG_EC_CONNECT(pg, win_title);
-
-    /* Misc */
-    PG_CONNECT(pg, on_Profile_Editor_destroy);
-    PG_CONNECT(pg, on_Profile_Editor_close);
-    PG_CONNECT(pg, on_Profile_Editor_response);
-    PG_CONNECT(pg, on_profile_notebook_switch_page);
-
-    /* Unique widgets */
-    PG_CONNECT(pg, on_font_set);
-    PG_CONNECT(pg, on_bgtype_toggled);
-    PG_CONNECT(pg, on_command_toggled);
-    PG_CONNECT(pg, on_close_buttons_toggled);
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
-    g_signal_connect(pg->bgimg, "file-set", G_CALLBACK(on_bgimg_chosen), pg);
-#else
-    PG_CONNECT(pg, on_bgimg_browse_clicked);
-#endif
-    PG_CONNECT(pg, on_reset_compat_clicked);
-#ifndef HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE
-    g_signal_connect(pg->cursor_blinks, "toggled",
-            G_CALLBACK(on_cursor_blinks_toggled), pg->profile);
-#endif
-    g_signal_connect(profilegui_widget(pg, "exit_action"), "changed",
-            G_CALLBACK(exit_action_changed), pg);
-    PG_CONNECT(pg, on_edit_ssh_clicked);
+    gtk_builder_connect_signals(pg->capp.builder, pg);
 }
 
 static void profilegui_fill_in_dialog(ProfileGUI * pg)
 {
-    GladeXML *glade = pg->glade;
-    Options *profile = pg->profile;
+    Options *profile = pg->capp.options;
     char *val;
 
     gtk_font_button_set_font_name(
@@ -620,140 +508,86 @@ static void profilegui_fill_in_dialog(ProfileGUI * pg)
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(profilegui_widget(pg,
                 "show_menubar")), !options_lookup_int_with_default(profile,
             "hide_menubar", FALSE));
-    capplet_set_boolean_toggle(glade, profile, "allow_bold", TRUE);
-    capplet_set_boolean_toggle(glade, profile, "audible_bell", TRUE);
-    capplet_set_boolean_toggle(glade, profile, "visible_bell", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "bell_highlights_tab", TRUE);
-#ifdef HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE
+    capplet_set_boolean_toggle(&pg->capp, "allow_bold", TRUE);
+    capplet_set_boolean_toggle(&pg->capp, "audible_bell", TRUE);
+    capplet_set_boolean_toggle(&pg->capp, "visible_bell", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "bell_highlights_tab", TRUE);
     {
         /* Use legacy cursor_blinks if cursor_blink_mode has not been set */
         int o = options_lookup_int(profile, "cursor_blinks") + 1;
         if (o)
             o ^= 3;
-        capplet_set_radio(glade, profile, "cursor_blink_mode", o);
+        capplet_set_radio(&pg->capp, "cursor_blink_mode", o);
     }
-#else
-    capplet_ignore_changes = TRUE;
-    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pg->cursor_blinks),
-            options_lookup_int_with_default(profile, "cursor_blinks", FALSE));
-    capplet_ignore_changes = FALSE;
-#endif
-#ifdef HAVE_VTE_TERMINAL_SET_CURSOR_SHAPE
-    capplet_set_radio(glade, profile, "cursor_shape", 0);
-#endif
-    capplet_set_boolean_toggle(glade, profile, "hide_mouse", TRUE);
-    capplet_set_text_entry(glade, profile, "sel_by_word", "-A-Za-z0-9,./?%&#_");
-    capplet_set_text_entry(glade, profile, "color_term", NULL);
-    capplet_set_text_entry(glade, profile, "term", NULL);
-    capplet_set_combo(glade, profile, "tab_pos", 0);
-    capplet_set_spin_button(glade, profile, "init_tabs", 1);
-    capplet_set_boolean_toggle(glade, profile, "wrap_switch_tab", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "tab_close_btn", TRUE);
-    capplet_set_boolean_toggle(glade, profile, "show_tab_status", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "always_show_tabs", TRUE);
+    capplet_set_radio(&pg->capp, "cursor_shape", 0);
+    capplet_set_boolean_toggle(&pg->capp, "hide_mouse", TRUE);
+    capplet_set_text_entry(&pg->capp, "sel_by_word", "-A-Za-z0-9,./?%&#_");
+    capplet_set_text_entry(&pg->capp, "color_term", NULL);
+    capplet_set_text_entry(&pg->capp, "term", NULL);
+    capplet_set_combo(&pg->capp, "tab_pos", 0);
+    capplet_set_spin_button(&pg->capp, "init_tabs", 1);
+    capplet_set_boolean_toggle(&pg->capp, "wrap_switch_tab", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "tab_close_btn", TRUE);
+    capplet_set_boolean_toggle(&pg->capp, "show_tab_status", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "always_show_tabs", TRUE);
+    capplet_set_boolean_toggle(&pg->capp, "new_tabs_adjacent", FALSE);
     profilegui_set_close_buttons_shading(pg);
-    capplet_set_text_entry(glade, profile, "browser", NULL);
-    capplet_set_radio(glade, profile, "browser_spawn_type", 0);
-    capplet_set_text_entry(glade, profile, "mailer", NULL);
-    capplet_set_radio(glade, profile, "mailer_spawn_type", 0);
-    capplet_set_spin_button(glade, profile, "width", 80);
-    capplet_set_spin_button(glade, profile, "height", 25);
-    capplet_set_radio(glade, profile, "background_type", 0);
+    capplet_set_text_entry(&pg->capp, "browser", NULL);
+    capplet_set_radio(&pg->capp, "browser_spawn_type", 0);
+    capplet_set_text_entry(&pg->capp, "mailer", NULL);
+    capplet_set_radio(&pg->capp, "mailer_spawn_type", 0);
+    capplet_set_text_entry(&pg->capp, "filer", NULL);
+    capplet_set_radio(&pg->capp, "filer_spawn_type", 0);
+    capplet_set_text_entry(&pg->capp, "dir_filer", NULL);
+    capplet_set_radio(&pg->capp, "dir_spawn_type", 0);
+    capplet_set_boolean_toggle(&pg->capp, "file_as_dir", TRUE);
+    capplet_set_boolean_toggle(&pg->capp, "match_plain_files", FALSE);
+    capplet_set_spin_button(&pg->capp, "width", 80);
+    capplet_set_spin_button(&pg->capp, "height", 25);
+    capplet_set_radio(&pg->capp, "background_type", 0);
+    on_cell_size_toggled(GTK_TOGGLE_BUTTON(profilegui_widget(pg, "cell_size")),
+            pg);
     profilegui_set_background_shading(pg);
-    capplet_set_float_range(glade, profile, "saturation", 1.0);
-    capplet_set_boolean_toggle(glade, profile, "scroll_background", TRUE);
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
+    capplet_set_float_range(&pg->capp, "saturation", 1.0);
+    capplet_set_boolean_toggle(&pg->capp, "scroll_background", TRUE);
     val = options_lookup_string_with_default(profile, "background_img", "");
-    profilegui_fill_in_file_chooser(pg, GTK_FILE_CHOOSER(pg->bgimg), val);
+    profilegui_fill_in_file_chooser(pg, val);
     g_free(val);
-#else
-    capplet_set_text_entry(glade, profile, "background_img", NULL);
-#endif
-    capplet_set_boolean_toggle(glade, profile, "true_translucence", TRUE);
-    capplet_set_radio(glade, profile, "scrollbar_pos", 1);
-    capplet_set_spin_button(glade, profile, "scrollback_lines", 1000);
-    capplet_set_boolean_toggle(glade, profile, "scroll_on_output", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "scroll_on_keystroke", FALSE);
-    capplet_set_combo(glade, profile, "backspace_binding",
+    capplet_set_radio(&pg->capp, "scrollbar_pos", 1);
+    capplet_set_spin_button(&pg->capp, "scrollback_lines", 1000);
+    capplet_set_boolean_toggle(&pg->capp, "scroll_on_output", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "scroll_on_keystroke", FALSE);
+    capplet_set_combo(&pg->capp, "backspace_binding",
             DEFAULT_BACKSPACE_BINDING);
-    capplet_set_combo(glade, profile, "delete_binding",
+    capplet_set_combo(&pg->capp, "delete_binding",
             DEFAULT_DELETE_BINDING);
-    capplet_set_boolean_toggle(glade, profile, "disable_menu_access", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "disable_menu_shortcuts", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "disable_tab_menu_shortcuts",
+    capplet_set_boolean_toggle(&pg->capp, "disable_menu_access", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "disable_menu_shortcuts", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "disable_tab_menu_shortcuts",
             FALSE);
     capplet_ignore_changes = TRUE;
-    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(glade,
-                "use_default_shell")), TRUE);
-    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(glade, "ssh_host")),
+    gtk_toggle_button_set_active(
+            GTK_TOGGLE_BUTTON(profilegui_widget(pg, "use_default_shell")),
+            TRUE);
+    gtk_entry_set_text(GTK_ENTRY(profilegui_widget(pg, "ssh_host")),
             options_lookup_string_with_default(profile, "ssh_address",
                     "localhost"));
     capplet_ignore_changes = FALSE;
-    capplet_set_boolean_toggle(glade, profile, "use_ssh", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "login_shell", FALSE);
-    capplet_set_boolean_toggle(glade, profile, "update_records", TRUE);
-    capplet_set_boolean_toggle(glade, profile, "use_custom_command", FALSE);
-    capplet_set_text_entry(glade, profile, "command", NULL);
-    capplet_set_combo(glade, profile, "exit_action", 0);
-    capplet_set_spin_button_float(glade, profile, "exit_pause");
-    capplet_set_text_entry(glade, profile, "title_string", "%s");
-    capplet_set_text_entry(glade, profile, "win_title", "%s");
-    exit_action_changed(GTK_COMBO_BOX(profilegui_widget(pg, "exit_action")),
-            pg);
+    capplet_set_boolean_toggle(&pg->capp, "use_ssh", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "login_shell", FALSE);
+    capplet_set_boolean_toggle(&pg->capp, "update_records", TRUE);
+    capplet_set_boolean_toggle(&pg->capp, "use_custom_command", FALSE);
+    capplet_set_text_entry(&pg->capp, "command", NULL);
+    capplet_set_combo(&pg->capp, "exit_action", 0);
+    capplet_set_spin_button_float(&pg->capp, "exit_pause");
+    capplet_set_text_entry(&pg->capp, "title_string", "%s");
+    capplet_set_text_entry(&pg->capp, "win_title", "%s");
+    exit_action_changed(
+        GTK_COMBO_BOX(capplet_lookup_combo(pg->capp.builder, "exit_action")),
+        pg);
     profilegui_set_command_shading(pg);
 }
 
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
-/* For backwards compatibility with older GTKs the glade file uses a text entry
- * and a button for choosing a background image file. This function replaces
- * them with a GtkFileChooserButton. */
-static void profilegui_replace_file_chooser_widgets(ProfileGUI *pg)
-{
-    GtkWidget *hbox = profilegui_widget(pg, "background_img_box");
-    GtkWidget *entry = profilegui_widget(pg, "background_img");
-    GtkWidget *button = profilegui_widget(pg, "bgimg_browse");
-
-    gtk_widget_destroy(entry);
-    gtk_widget_destroy(button);
-    pg->bgimg = gtk_file_chooser_button_new(_("Background image"),
-            GTK_FILE_CHOOSER_ACTION_OPEN);
-    profilegui_setup_file_chooser(pg, GTK_FILE_CHOOSER(pg->bgimg));
-    gtk_widget_show(pg->bgimg);
-    gtk_box_pack_start(GTK_BOX(hbox), pg->bgimg, TRUE, TRUE, 0);
-}
-#endif
-
-#if !defined HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE || \
-        !defined HAVE_VTE_TERMINAL_SET_CURSOR_SHAPE
-static void profilegui_destroy_radio_group_and_label(ProfileGUI *pg,
-        const char *group_name)
-{
-    int n;
-    char *name = g_strdup_printf("%s_label", group_name);
-    
-    gtk_widget_destroy(profilegui_widget(pg, name));
-    g_free(name);
-    for (n = 0; n < 3; ++n)
-    {
-        name = g_strdup_printf("%s%d", group_name, n);
-        gtk_widget_destroy(profilegui_widget(pg, name));
-        g_free(name);
-    }
-}
-#endif
-
-#ifndef HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE
-static void profilegui_replace_blink_mode_widgets(ProfileGUI *pg)
-{
-    GtkWidget *table = profilegui_widget(pg, "general_table");
-    
-    profilegui_destroy_radio_group_and_label(pg, "cursor_blink_mode");
-    pg->cursor_blinks = gtk_check_button_new_with_mnemonic(_("Cursor blin_ks"));
-    gtk_widget_show(pg->cursor_blinks);
-    gtk_table_attach_defaults(GTK_TABLE(table), pg->cursor_blinks, 2, 3, 2, 3);
-}
-#endif
-
 static gboolean page_selected(GtkTreeSelection *selection,
         GtkTreeModel *model, GtkTreePath *path,
         gboolean path_currently_selected, gpointer pg)
@@ -773,7 +607,7 @@ static void profilegui_setup_list_store(ProfileGUI *pg)
 {
     static char const *labels[] = {
             N_("Appearance"), N_("General"), N_("Command"),
-            N_("URI handling"), N_("Background"), N_("Scrolling"),
+            N_("Net URIs"), N_("File URIs"), N_("Background"), N_("Scrolling"),
             N_("Keyboard"), N_("Tabs")
     };
     GtkTreeIter iter;
@@ -784,7 +618,7 @@ static void profilegui_setup_list_store(ProfileGUI *pg)
     GtkTreeSelection *sel;
     
     pg->list_store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_INT);
-    for (n = 0; n < 8; ++n)
+    for (n = 0; n < G_N_ELEMENTS(labels); ++n)
     {
         gtk_list_store_append(pg->list_store, &iter);
         gtk_list_store_set(pg->list_store, &iter,
@@ -805,12 +639,109 @@ static void profilegui_setup_list_store(ProfileGUI *pg)
     gtk_widget_show(tvw);
 }
 
+/* Hopefully all this combo support stuff can go away when everything
+ * supports GtkComboBoxText.
+ */
+
+static void profilegui_add_combo_items(GtkWidget *combo, ...)
+{
+    const char *item;
+    va_list ap;
+#ifdef HAVE_GTK_COMBO_BOX_TEXT_NEW
+    GtkComboBoxText *cast_combo = GTK_COMBO_BOX_TEXT(combo);
+#else
+    GtkComboBox *cast_combo = GTK_COMBO_BOX(combo);
+#endif
+    
+    va_start(ap, combo);
+    while ((item = va_arg(ap, const char *)) != NULL)
+    {
+#ifdef HAVE_GTK_COMBO_BOX_TEXT_NEW
+        gtk_combo_box_text_append_text(cast_combo, item);
+#else
+        gtk_combo_box_append_text(cast_combo, item);
+#endif
+    }
+    va_end(ap);
+}
+
+static void profilegui_add_combo_items_by_name(GtkWidget *combo,
+        const char *name)
+{
+    if (!strcmp(name, "exit_action"))
+    {
+        profilegui_add_combo_items(combo,
+            _("Exit terminal"),
+            _("Hold terminal open"),
+            _("Restart command"),
+            _("Ask user"),
+            NULL);
+    }
+    else if (!strcmp(name, "delete_binding") ||
+            !strcmp(name, "backspace_binding"))
+    {
+        profilegui_add_combo_items(combo,
+            _("Automatic choice"),
+            _("Control-H"),
+            _("ASCII DEL"),
+            _("Escape sequence"),
+            NULL);
+    }
+    else if (!strcmp(name, "tab_pos"))
+    {
+        profilegui_add_combo_items(combo,
+            _("Top"),
+            _("Bottom"),
+            _("Left"),
+            _("Right"),
+            _("Hidden"),
+            NULL);
+    }
+}
+
+static void profilegui_make_a_combo(ProfileGUI *pg, const char *name)
+{
+    char *box_name = g_strdup_printf("%s_hbox", name);
+    GtkBox *box = GTK_BOX(gtk_builder_get_object(pg->capp.builder, box_name));
+    GtkWidget *combo;
+    
+#ifdef HAVE_GTK_COMBO_BOX_TEXT_NEW
+    combo = gtk_combo_box_text_new();
+#else
+    combo = gtk_combo_box_new_text();
+#endif
+    profilegui_add_combo_items_by_name(combo, name);
+    gtk_buildable_set_name(GTK_BUILDABLE(combo), name);
+    gtk_buildable_add_child(GTK_BUILDABLE(box), pg->capp.builder,
+            G_OBJECT(combo), NULL);
+    g_signal_connect(combo, "changed", G_CALLBACK(on_combo_changed), pg);
+    gtk_widget_show(combo);
+}
+
+static void profilegui_make_combos(ProfileGUI *pg)
+{
+    profilegui_make_a_combo(pg, "exit_action");
+    profilegui_make_a_combo(pg, "backspace_binding");
+    profilegui_make_a_combo(pg, "delete_binding");
+    profilegui_make_a_combo(pg, "tab_pos");
+}
+
 /* Loads a Profile and creates a working dialog box for it */
 ProfileGUI *profilegui_open(const char *profile_name, GdkScreen *scrn)
 {
     static DynamicOptions *profiles = NULL;
     ProfileGUI *pg;
     char *title;
+    static const char *adj_names[] = {
+            "width_adjustment", "height_adjustment",
+            "exit_pause_adjustment", "scrollback_lines_adjustment",
+            "init_tabs_adjustment", "saturation_adjustment",
+            NULL };
+    static const char *obj_names[] = {
+            "Profile_Editor", "ssh_dialog",
+            "general_entries_size_group", "general_entry_labels_size_group", 
+            NULL };
+    GError *error = NULL;
 
     if (!profilegui_being_edited)
     {
@@ -821,8 +752,8 @@ ProfileGUI *profilegui_open(const char *profile_name, GdkScreen *scrn)
     if (pg)
     {
         if (scrn)
-            gtk_window_set_screen(GTK_WINDOW(pg->widget), scrn);
-        gtk_window_present(GTK_WINDOW(pg->widget));
+            gtk_window_set_screen(GTK_WINDOW(pg->dialog), scrn);
+        gtk_window_present(GTK_WINDOW(pg->dialog));
         return pg;
     }
     
@@ -831,42 +762,41 @@ ProfileGUI *profilegui_open(const char *profile_name, GdkScreen *scrn)
 
     pg = g_new0(ProfileGUI, 1);
     pg->profile_name = g_strdup(profile_name);
-    pg->profile = dynamic_options_lookup_and_ref(profiles, profile_name,
+    pg->capp.options = dynamic_options_lookup_and_ref(profiles, profile_name,
         "roxterm profile");
 
-    pg->glade = glade_xml_new(capplet_get_glade_filename(),
-        "Profile_Editor", NULL);
-    g_assert(pg->glade);
-    pg->widget = profilegui_widget(pg, "Profile_Editor");
+    pg->capp.builder = gtk_builder_new();
+    if (!gtk_builder_add_objects_from_file(pg->capp.builder, 
+            capplet_get_ui_filename(), (char **) adj_names, &error) ||
+            !gtk_builder_add_objects_from_file(pg->capp.builder, 
+            capplet_get_ui_filename(), (char **) obj_names, &error))
+    {
+        g_error(_("Unable to load GTK UI definitions: %s"), error->message);
+    }
+    pg->dialog = profilegui_widget(pg, "Profile_Editor");
+    pg->ssh_dialog = profilegui_widget(pg, "ssh_dialog");
+    
     if (scrn)
-        gtk_window_set_screen(GTK_WINDOW(pg->widget), scrn);
+        gtk_window_set_screen(GTK_WINDOW(pg->dialog), scrn);
 
     title = g_strdup_printf(_("ROXTerm Profile \"%s\""), profile_name);
-    gtk_window_set_title(GTK_WINDOW(pg->widget), title);
+    gtk_window_set_title(GTK_WINDOW(pg->dialog), title);
     g_free(title);
     
+    profilegui_make_combos(pg);
+    
     profilegui_setup_list_store(pg);
 
     g_hash_table_insert(profilegui_being_edited, g_strdup(profile_name), pg);
 
-#ifndef HAVE_VTE_TERMINAL_SET_CURSOR_BLINK_MODE
-    profilegui_replace_blink_mode_widgets(pg);
-#endif
-#ifndef HAVE_VTE_TERMINAL_SET_CURSOR_SHAPE
-    profilegui_destroy_radio_group_and_label(pg, "cursor_shape");
-#endif
-#ifdef HAVE_GTK_FILE_CHOOSER_BUTTON_NEW
-    profilegui_replace_file_chooser_widgets(pg);
-#else
-    pg->bgimg = profilegui_widget(pg, "background_img");
-#endif
+    profilegui_setup_file_chooser(pg);
     pg->bgimg_drd = drag_receive_setup_dest_widget(
-            profilegui_widget(pg, "bgimg_drag_target"),
+            profilegui_widget(pg, "bgimg_drag_target_vbox"),
             bgimg_drag_data_received, NULL, pg);
     profilegui_fill_in_dialog(pg);
     profilegui_connect_handlers(pg);
 
-    gtk_widget_show(pg->widget);
+    gtk_widget_show(pg->dialog);
 
     capplet_inc_windows();
     configlet_lock_profiles();
@@ -881,6 +811,7 @@ void profilegui_delete(ProfileGUI * pg)
     if (!dynopts)
         dynopts = dynamic_options_get("Profiles");
 
+    profilegui_check_ssh_entries_for_changes(pg);
     profilegui_check_entries_for_changes(pg);
     if (!g_hash_table_remove(profilegui_being_edited, pg->profile_name))
     {
@@ -890,10 +821,11 @@ void profilegui_delete(ProfileGUI * pg)
     if (!pg->ignore_destroy)
     {
         pg->ignore_destroy = TRUE;
-        gtk_widget_destroy(profilegui_widget(pg, "Profile_Editor"));
+        gtk_widget_destroy(pg->dialog);
+        gtk_widget_destroy(pg->ssh_dialog);
     }
     UNREF_LOG(g_object_unref(pg->list_store));
-    UNREF_LOG(g_object_unref(pg->glade));
+    UNREF_LOG(g_object_unref(pg->capp.builder));
     dynamic_options_unref(dynopts, pg->profile_name);
     g_free(pg->profile_name);
     drag_receive_data_delete(pg->bgimg_drd);
